--=============================================================================
-- Description: IBIS macro library implemented in Verilog-A(MS)
-- Created: October 2005 by Arpad Muranyi (Intel Corporation)
--
-- Revision history:
--
-- 2005/10/10  Verilog-A rev0.0 completed
-- 2005/10/19  - Fixed IBIS_K to include the Scale parameter in the
--               calculation of M.
--             - Fixed IBIS_T equations
-- 2005/10/27  - Added the 10 PWL sources
-- 2005/11/03  - Changed IBIS_TCVS_PWL and IBIS_TCCS_PWL sources to do HE
--             - Changed the four event controlled PWL sources to have two
--               data tables with independent threshold voltages (one for
--               each triggering edge direction), and to do HE
--
-- =========================================modulelist=====
--
-- LIST OF MODULES:
--
--   IBIS_VCCS_MAX        (p, n, ps1, ns1, ps2, ns2)
--   IBIS_VCVS_MULT       (p, n, ps1, ns1, ps2, ns2)
--   IBIS_VCVS_DELAY      (p, n, ps, ns)
--   IBIS_VCVS            (p, n, ps, ns)
--   IBIS_CCVS_SUM        (p, n, ps1, ns1, ps2, ns2)
--   IBIS_V               (p, n)
--   IBIS_VCC             (p, n, ps, ns)
--   IBIS_VCCS_DIV        (p, n, ps1, ns1, ps2, ns2)
--   IBIS_OPENSINK        (PU_ref, PD_ref, IO, In_D, PC_ref, GC_ref)
--   IBIS_VECCS_PWL       (p, n, ps, ns)
--   IBIS_CCCS_MAX        (p, n, ps1, ns1, ps2, ns2)
--   IBIS_CCVS_PWL        (p, n, ps, ns)
--   IBIS_VCCS_MULT       (p, n, ps1, ns1, ps2, ns2)
--   IBIS_I               (p, n)
--   IBIS_L               (p, n)
--   IBIS_CECCS_PWL       (p, n, ps, ns)
--   IBIS_CCCS_DIV        (p, n, ps1, ns1, ps2, ns2)
--   IBIS_CCVS_DIV        (p, n, ps1, ns1, ps2, ns2)
--   IBIS_CCCS_PWL        (p, n, ps, ns)
--   IBIS_3STATE          (PU_ref, PD_ref, IO, In_D, En_D, PC_ref, GC_ref)
--   IBIS_CCCS_SUM        (p, n, ps1, ns1, ps2, ns2)
--   IBIS_INPUT           (PC_ref, GC_ref, Input, Rcv_D)
--   IBIS_VCCS_DELAY      (p, n, ps, ns)
--   IBIS_TCCS_PWL        (p, n)
--   IBIS_CCVS_MAX        (p, n, ps1, ns1, ps2, ns2)
--   IBIS_VCCS_SUM        (p, n, ps1, ns1, ps2, ns2)
--   IBIS_IO_OPENSOURCE   (PU_ref, PD_ref, IO, In_D, En_D, Rcv_D, PC_ref, GC_ref)
--   IBIS_VCL             (p, n, ps, ns)
--   IBIS_VCVS_MIN        (p, n, ps1, ns1, ps2, ns2)
--   IBIS_IO_OPENSINK     (PU_ref, PD_ref, IO, In_D, En_D, Rcv_D, PC_ref, GC_ref)
--   IBIS_CCCS_ABS        (p, n, ps, ns)
--   IBIS_K               (p1, n1, p2, n2)
--   IBIS_T               (n1, ref1, n2, ref2)
--   IBIS_VCCS_MIN        (p, n, ps1, ns1, ps2, ns2)
--   IBIS_CCVS_ABS        (p, n, ps, ns)
--   IBIS_CCVS_MULT       (p, n, ps1, ns1, ps2, ns2)
--   IBIS_OPENSOURCE      (PU_ref, PD_ref, IO, In_D, PC_ref, GC_ref)
--   IBIS_CCCS            (p, n, ps, ns)
--   IBIS_TCVS_PWL        (p, n)
--   IBIS_IO              (PU_ref, PD_ref, IO, In_D, En_D, Rcv_D, PC_ref, GC_ref)
--   IBIS_R               (p, n)
--   IBIS_VCVS_SUM        (p, n, ps1, ns1, ps2, ns2)
--   IBIS_OUTPUT          (PU_ref, PD_ref, Output, In_D, PC_ref, GC_ref)
--   IBIS_CCC             (p, n, ps, ns)
--   IBIS_VCVS_MAX        (p, n, ps1, ns1, ps2, ns2)
--   IBIS_VCCS_PWL        (p, n, ps, ns)
--   IBIS_VCVS_DIV        (p, n, ps1, ns1, ps2, ns2)
--   IBIS_CCR             (p, n, ps, ns)
--   IBIS_VCCS            (p, n, ps, ns)
--   IBIS_CCCS_MULT       (p, n, ps1, ns1, ps2, ns2)
--   IBIS_CCCS_MIN        (p, n, ps1, ns1, ps2, ns2)
--   IBIS_CCCS_DELAY      (p, n, ps, ns)
--   IBIS_CECVS_PWL       (p, n, ps, ns)
--   IBIS_CCVS            (p, n, ps, ns)
--   IBIS_CCL             (p, n, ps, ns)
--   IBIS_VCCS_ABS        (p, n, ps, ns)
--   IBIS_C               (p, n)
--   IBIS_CCVS_MIN        (p, n, ps1, ns1, ps2, ns2)
--   IBIS_CCVS_DELAY      (p, n, ps, ns)
--   IBIS_VCVS_ABS        (p, n, ps, ns)
--   IBIS_VCVS_PWL        (p, n, ps, ns)
--   IBIS_VECVS_PWL       (p, n, ps, ns)
--   IBIS_VCR             (p, n, ps, ns)
--
-- =========================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_R (p, n)
--
-- DESCRIPTION:
--              A simple resistor. The resistance value is calculated:
--
--              R = Rval * Scale
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--
-- PARAMETERS:
--
--  Rval        Resistance value
--              DEFAULT Rval = '1.0'
--
--  Scale       Scaling factor for calculating resistance value
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 clk_n clkp IBIS_R Rval=100
--
--  VAMS:       IBIS_R #(.Rval(100)) R1 (clk_n, clk_p);
--
-- =========================================================
  entity IBIS_R is

    generic ( 
      Rval  : real := 1.0;
      Scale : real := 1.0);

    port (terminal p, n : electrical);

  end entity IBIS_R;
  --===========================================================================
  architecture ideal of IBIS_R is

  quantity Vout across Iout through p to n;

  begin

    Vout == Scale * Rval * Iout;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_VCR (p, n, ps, ns)
--
-- DESCRIPTION:
--              A voltage-controlled resistor. The resistance value is
--              calculated:
--
--              R = V(ps,ns) * Scale
--
--              The sense terminals ps and ns draw no current.
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--  ps          positive sense terminal for control voltage
--  ns          negative sense terminal for control voltage
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating resistance value
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 clkn clkp vcrp vcrn IBIS_VCR Scale=50
--
--  VAMS:       IBIS_VCR #(.Scale(50)) VCR1 (clkn, clk_p, vcrp, vcrn);
--
-- =========================================================
  entity IBIS_VCR is

    generic ( 
      Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_VCR;
  --===========================================================================
  architecture ideal of IBIS_VCR is

  quantity Vout across Iout through p  to n;
  quantity Vin  across              ps to ns;

  begin

    Vout == Scale * Vin * Iout;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_CCR (p, n, ps, ns)
--
-- DESCRIPTION:
--              A current-controlled resistor. The resistance value is
--              calculated:
--
--              R = I(ps,ns) * Scale
--
--              The sense terminals ps and ns behave as a short circuit.
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--  ps          positive sense terminal for control current
--  ns          negative sense terminal for control current
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating resistance value
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 clkn clkp ccrp ccrn IBIS_CCR Scale=200
--
--  VAMS:       IBIS_CCR #(.Scale(200)) CCR1 (clkn, clk_p, ccrp, ccrn);
--
-- =========================================================
  entity IBIS_CCR is

    generic ( 
      Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_CCR;
  --===========================================================================
  architecture ideal of IBIS_CCR is

  quantity Vout across Iout through p  to n;
  quantity Vin  across Iin  through ps to ns;

  begin

    Vin  == 0.0;
    Vout == Scale * Iin * Iout;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_C (p, n)
--
-- DESCRIPTION:
--              A simple capacitor. The capacitance value is calculated:
--
--              C = Cval * Scale
--
-- TERMINALS:
--
--  p           positive capacitor terminal
--  n           negative capacitor terminal
--
-- PARAMETERS:
--
--  Cval        Capacitance value
--              DEFAULT Cval = '1.0'
--
--  V0          Voltage value at time = zero
--              DEFAULT V0 = '0.0'
--
--  Scale       Scaling factor for calculating capacitance value
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 clk_n clkp IBIS_C Cval=4.5 Scale=1e-12
--
--  VAMS:       IBIS_C #(.Cval(4.5),.Scale(1e-12))) C1 (clk_n, clk_p);
--
-- =========================================================
  entity IBIS_C is

    generic ( 
      Cval  : real := 1.0;
      V0    : real := 0.0;
      Scale : real := 1.0);

    port (terminal p, n : electrical);

  end entity IBIS_C;
  --===========================================================================
  architecture ideal of IBIS_C is

  quantity Vout across Iout through p to n;

  begin

    if domain = quiescent_domain use
      Vout == V0;
    else
      Iout == Scale * Cval * Vout'dot;
    end use;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_VCC (p, n, ps, ns)
--
-- DESCRIPTION:
--              A voltage-controlled capacitor. The capacitance value is
--              calculated:
--
--              C = V(ps,ns) * Scale
--
--              The sense terminals ps and ns draw no current. This is a
--              charge-conserving model. When the capacitance value
--              changes, a displacement current is produced from p to n to
--              adjust the stored charge so that the total charge level
--              observes the rule:
--
--              Q = C * V(p,n)
--
-- TERMINALS:
--
--  p           positive capacitor terminal
--  n           negative capacitor terminal
--  ps          positive sense terminal for control voltage
--  ns          negative sense terminal for control voltage
--
-- PARAMETERS:
--
--  V0          Voltage at time = zero
--              DEFAULT V0 = '0.0'
--
--  Scale       Scaling factor for calculating capacitance value
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 clkn clkp vccp vccn IBIS_VCC Scale=1e-12
--
--  VAMS:       IBIS_VCC #(.Scale(1e-12)) VCC1 (clkn, clk_p, vccp, vccn);
--
-- =========================================================
  entity IBIS_VCC is

    generic ( 
      V0    : real := 0.0;
      Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_VCC;
  --===========================================================================
  architecture ideal of IBIS_VCC is

  quantity Vout across Iout through p  to n;
  quantity Vin  across              ps to ns;

  begin

    if domain = quiescent_domain use
      Vout == V0;
    else
      Iout == Vout * Scale * Vin'dot + Scale * Vin * Vout'dot;
    end use;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_CCC (p, n, ps, ns)
--
-- DESCRIPTION:
--              A current-controlled capacitor. The capacitance value is
--              calculated:
--
--              C = I(ps,ns) * Scale
--
--              The sense terminals ps and ns behave as a short circuit.
--
--              This is a charge-conserving model. When the capacitance
--              value changes, a displacement current is produced from p
--              to n to adjust the stored charge so that the total charge
--              level observes the rule:
--
--              Q = C * V(p,n)
--
-- TERMINALS:
--
--  p           positive capacitor terminal
--  n           negative capacitor terminal
--  ps          positive sense terminal for control current
--  ns          negative sense terminal for control current
--
-- PARAMETERS:
--
--  V0          Capacitor voltage at time zero
--              DEFAULT V0 = '0.0'
--
--  Scale       Scaling factor for calculating capacitance value
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 clkn clkp cccp cccn IBIS_CCC Scale=5e-12
--
--  VAMS:       IBIS_CCC #(.Scale(5e-12)) CCC1 (clkn, clk_p, cccp, cccn);
--
-- =========================================================
  entity IBIS_CCC is

    generic ( 
      V0    : real := 0.0;
      Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_CCC;
  --===========================================================================
  architecture ideal of IBIS_CCC is

  quantity Vout across Iout through p  to n;
  quantity Vin  across Iin  through ps to ns;

  begin

    Vin  == 0.0;
    if domain = quiescent_domain use
      Vout == V0;
    else
      Iout == Vout * Scale * Iin'dot + Scale * Iin * Vout'dot;
    end use;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_L (p, n)
--
-- DESCRIPTION:
--              A simple inductor. The inductance value is calculated:
--
--              L = Lval * Scale
--
-- TERMINALS:
--
--  p           positive inductor terminal
--  n           negative inductor terminal
--
-- PARAMETERS:
--
--  Lval        Inductance value
--              DEFAULT Lval = '1.0'
--
--  I0          Inductance value
--              DEFAULT I0 = '0.0'
--
--  Scale       Scaling factor for calculating inductance value
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 clk_n clkp IBIS_L Lval=0.8 Scale=1e-9
--
--  VAMS:       IBIS_L #(.Lval(0.8),.Scale(1e-9)) L1 (clk_n, clk_p);
--
-- =========================================================
  entity IBIS_L is

    generic ( 
      Lval  : real := 1.0;
      I0    : real := 0.0;
      Scale : real := 1.0);

    port (terminal p, n : electrical);

  end entity IBIS_L;
  --===========================================================================
  architecture ideal of IBIS_L is

  quantity Vout across Iout through p to n;

  begin

    if domain = quiescent_domain use
      Iout == I0;
    else
      Vout == Scale * Lval * Iout'dot;
    end use;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_VCL (p, n, ps, ns)
--
-- DESCRIPTION:
--              A voltage-controlled inductor. The inductance value is
--              calculated:
--
--              L = V(ps,ns) * Scale
--
--              The sense terminals ps and ns draw no current. This is a
--              flux-conserving model. When the inductance value changes,
--              the voltage-current relationship adjusts so that magnetic
--              flux is conserved in the equation:
--
--              B = L * I(p,n)
--
-- TERMINALS:
--
--  p           positive inductor terminal
--  n           negative inductor terminal
--  ps          positive sense terminal for control voltage
--  ns          negative sense terminal for control voltage
--
-- PARAMETERS:
--
--  I0          Inductor current at time zero
--              DEFAULT I0 = '0.0'
--
--  Scale       Scaling factor for calculating inductance value
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 clkn clkp vclp vcln IBIS_VCL Scale=0.5e-9
--
--  VAMS:       IBIS_VCL #(.Scale(0.5e-9)) VCL1 (clkn, clk_p, vclp, vcln);
--
-- =========================================================
  entity IBIS_VCL is

    generic ( 
      I0    : real := 0.0;
      Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_VCL;
  --===========================================================================
  architecture ideal of IBIS_VCL is

  quantity Vout across Iout through p  to n;
  quantity Vin  across              ps to ns;

  begin

    if domain = quiescent_domain use
      Iout == I0;
    else
      Vout == Iout * Scale * Vin'dot + Scale * Vin * Iout'dot;
    end use;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_CCL (p, n, ps, ns)
--
-- DESCRIPTION:
--              A current-controlled inductor. The inductance value is
--              calculated:
--
--              L = I(ps,ns) * Scale
--
--              The sense terminals ps and ns draw no current.
--
--              This is a flux-conserving model. When the inductance value
--              changes, the voltage-current relationship adjusts so that
--              magnetic flux is conserved in the equation:
--
--              B = L * I(p,n)
--
-- TERMINALS:
--
--  p           positive inductor terminal
--  n           negative inductor terminal
--  ps          positive sense terminal for control current
--  ns          negative sense terminal for control current
--
-- PARAMETERS:
--
--  I0          Inductor current at time zero
--              DEFAULT I0 = '0.0'
--
--  Scale       Scaling factor for calculating inductance value
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 clkn clkp cclp ccln IBIS_CCL Scale=4e-9
--
--  VAMS:       IBIS_CCL #(.Scale(4e-9)) CCL1 (clkn, clk_p, cclp, ccln);
--
-- =========================================================
  entity IBIS_CCL is

    generic ( 
      I0    : real := 0.0;
      Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_CCL;
  --===========================================================================
  architecture ideal of IBIS_CCL is

  quantity Vout across Iout through p  to n;
  quantity Vin  across Iin  through ps to ns;

  begin

    Vin  == 0.0;
    if domain = quiescent_domain use
      Iout == I0;
    else
      Vout == Iout * Scale * Iin'dot + Scale * Iin * Iout'dot;
    end use;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_K (p1, n1, p2, n2)
--
-- DESCRIPTION:
--              A mutually-coupled inductor pair. The inductance values
--              are calculated:
--
--              L_1 = Lval_1 * Scale
--
--              L_2 = Lval_2 * Scale
--
-- TERMINALS:
--
--  p1          positive inductor 1 terminal
--  n1          negative inductor 1 terminal
--  p2          positive inductor 2 terminal
--  n2          negative inductor 2 terminal
--
-- PARAMETERS:
--
--  Lval_1      Inductance value for inductor 1
--              DEFAULT Lval_1 = '1.0'
--
--  Lval_2      Inductance value for inductor 2
--              DEFAULT Lval_2 = '1.0'
--
--  Kval        Mutual coupling value
--              DEFAULT Kval = '0.0'
--
--  I0_1        Inductance value at time zero for inductor 1
--              DEFAULT I0_1 = '0.0'
--
--  I0_2        Inductance value at time zero for inductor 2
--              DEFAULT I0_2 = '0.0'
--
--  Scale       Scaling factor for calculating both inductance values
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 pad_n pad_p pin_n pin_p IBIS_K \
--              + Lval_1=0.8 Lval_2=0.5 Kval=0.15 Scale=1e-9
--
--  VAMS:       IBIS_K
--              #(.Lval_1(0.8),.Lval_2(0.5),.Kval(0.15),.Scale(1e-9)) \
--              K1 (pad_n, pad_p, pin_n, pin_p);
--
-- =========================================================
  entity IBIS_K is

    generic ( 
      Lval_1 : real := 1.0;
      Lval_2 : real := 1.0;
      Kval   : real := 0.0;
      I0_1   : real := 0.0;
      I0_2   : real := 0.0;
      Scale  : real := 1.0);

    port (terminal p1, n1, p2, n2 : electrical);

  end entity IBIS_K;
  --===========================================================================
  architecture ideal of IBIS_K is

  quantity Vout1 across Iout1 through p1 to n1;
  quantity Vout2 across Iout2 through p2 to n2;
  constant M : real := Scale * Kval * sqrt(Lval_1 * Lval_2);

  begin

    if domain = quiescent_domain use
      Iout1 == I0_1;
      Iout2 == I0_2;
    else
      Vout1 == Scale * Lval_1 * Iout1'dot + M * Iout2'dot;
      Vout2 == Scale * Lval_2 * Iout2'dot + M * Iout1'dot;
    end use;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_V (p, n)
--
-- DESCRIPTION:
--              A DC voltage source. Output voltage is calculated:
--
--              V(p,n) = Vdc * Scale
--
-- TERMINALS:
--
--  p           positive voltage source terminal
--  n           negative voltage source terminal
--
-- PARAMETERS:
--
--  Vdc         DC voltage value
--              DEFAULT Vdc = '1.0'
--
--  Scale       Scaling factor for calculating inductance value
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 vddq vss IBIS_V Vdc=3.3
--
--  VAMS:       IBIS_V #(.Vdc(3.3)) V1 (vddq, vss);
--
-- =========================================================
  entity IBIS_V is

    generic ( 
      Vdc   : real := 1.0;
      Scale : real := 1.0);

    port (terminal p, n : electrical);

  end entity IBIS_V;
  --===========================================================================
  architecture ideal of IBIS_V is

  quantity Vout across Iout through p to n;

  begin

    Vout == Scale * Vdc;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_VCVS (p, n, ps, ns)
--
-- DESCRIPTION:
--              A voltage-controlled voltage source. The output voltage
--              value is calculated:
--
--              V(p,n) = V(ps,ns) * Scale
--
--              The sense terminals ps and ns draw no current.
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--  ps          positive sense terminal for control voltage
--  ns          negative sense terminal for control voltage
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 clkn clkp vcrp vcrn IBIS_VCVS Scale=0.5
--
--  VAMS:       IBIS_VCVS #(.Scale(0.5)) VCVS1 (clkn, clk_p, vcrp, vcrn);
--
-- =========================================================
  entity IBIS_VCVS is

    generic ( 
      Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_VCVS;
  --===========================================================================
  architecture ideal of IBIS_VCVS is

  quantity Vout across Iout through p  to n;
  quantity Vin  across              ps to ns;

  begin

    Vout == Scale * Vin;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_CCVS (p, n, ps, ns)
--
-- DESCRIPTION:
--              A current-controlled voltage source. The output voltage
--              value is calculated:
--
--              V(p,n) = I(ps,ns) * Scale
--
--              The sense terminals ps and ns behave as a short circuit.
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--  ps          positive sense terminal for control current
--  ns          negative sense terminal for control current
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 buf2in vss pad padi IBIS_CCVS Scale=100
--
--  VAMS:       IBIS_CCVS #(.Scale(100)) CCVS1 (buf2in, vss, pad, padi);
--
-- =========================================================
  entity IBIS_CCVS is

    generic ( 
      Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_CCVS;
  --===========================================================================
  architecture ideal of IBIS_CCVS is

  quantity Vout across Iout through p  to n;
  quantity Vin  across Iin  through ps to ns;

  begin

    Vin  == 0.0;
    Vout == Scale * Iin;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_VCVS_DELAY (p, n, ps, ns)
--
-- DESCRIPTION:
--              A voltage-controlled voltage source with delay. All
--              changes at the sense terminals affect the output later in
--              time, using the specified delay. The voltage value is
--              calculated:
--
--              V(p,n) = V(ps,ns) * Scale, delayed by TD
--
--              The sense terminals ps and ns draw no current.
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--  ps          positive sense terminal for control voltage
--  ns          negative sense terminal for control voltage
--
-- PARAMETERS:
--
--  TD          Time delay from control voltage change to output change
--              DEFAULT TD = '0.0'
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 buf2in vss pad padi IBIS_VCVS_DELAY TD=200p
--
--  VAMS:       IBIS_VCVS_DELAY #(.TD(200e-12)) \
--              VCVS1 (buf2in, vss, pad, padi);
--
-- =========================================================
  entity IBIS_VCVS_DELAY is

    generic ( 
      TD    : real := 0.0;
      Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_VCVS_DELAY;
  --===========================================================================
  architecture ideal of IBIS_VCVS_DELAY is

  quantity Vout across Iout through p  to n;
  quantity Vin  across              ps to ns;

  begin

    Vout == Scale * Vin'delayed(TD);

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_CCVS_DELAY (p, n, ps, ns)
--
-- DESCRIPTION:
--              A current-controlled voltage source with delay. All
--              changes at the sense terminals affect the output later,
--              using the specified delay in seconds. The voltage value is
--              calculated:
--
--              V(p,n) = I(ps,ns) * Scale
--
--              The sense terminals ps and ns behave as a short circuit.
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--  ps          positive sense terminal for control current
--  ns          negative sense terminal for control current
--
-- PARAMETERS:
--
--  TD          Time delay from control current change to output change
--              DEFAULT TD = '0.0'
--
--  Scale       Scaling factor for calculating voltage
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 buf2in vss pad padi IBIS_CCVS_DELAY Scale=100 TD=200p
--
--  VAMS:       IBIS_CCVS_DELAY #(.Scale(100),.TD(200e-12)) \
--              CCVS1 (buf2in, vss, pad, padi);
--
-- =========================================================
  entity IBIS_CCVS_DELAY is

    generic ( 
      TD    : real := 0.0;
      Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_CCVS_DELAY;
  --===========================================================================
  architecture ideal of IBIS_CCVS_DELAY is

  quantity Vout across Iout through p  to n;
  quantity Vin  across Iin  through ps to ns;
  quantity Iin_as_real  : real;   -- Delete these two lines when the last line
  quantity Iin_real_del : real;   -- in the code works

  begin

    Vin  == 0.0;
    Iin_as_real  == Iin;
    Iin_real_del == Iin_as_real'delayed(TD);
    Vout == Scale * Iin_real_del;
--  Vout == Scale * Iin'delayed(TD);   -- Delete the above 3 lines if this works

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_VCVS_MIN (p, n, ps1, ns1, ps2, ns2)
--
-- DESCRIPTION:
--              A voltage-controlled voltage source for taking the lesser
--              of two control input voltages. The output voltage value is
--              calculated:
--
--              V(p,n) = min( V(ps1,ns1) V(ps2,ns2) ) * Scale
--
--              The sense terminals ps1 and ns1 draw no current. The sense
--              terminals ps2 and ns2 draw no current.
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--  ps1         positive sense terminal for control voltage 1
--  ns1         negative sense terminal for control voltage 1
--  ps2         positive sense terminal for control voltage 2
--  ns2         negative sense terminal for control voltage 2
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 bufctl vss ctlmain vss boost vss IBIS_VCVS_MIN
--
--  VAMS:       IBIS_VCVS_MIN MIN1 (bufctl, vss, ctlmain, vss, boost,
--              vss);
--
-- =========================================================
  entity IBIS_VCVS_MIN is

    generic ( 
      Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_VCVS_MIN;
  --===========================================================================
  architecture ideal of IBIS_VCVS_MIN is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across              ps1 to ns1;
  quantity Vin2 across              ps2 to ns2;

  begin

    Vout == Scale * realmin(Vin1, Vin2);

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_CCVS_MIN (p, n, ps1, ns1, ps2, ns2)
--
-- DESCRIPTION:
--              A voltage-controlled voltage source for taking the lesser
--              of two control input voltages. The output voltage value is
--              calculated:
--
--              V(p,n) = min( V(ps1,ns1) V(ps2,ns2) ) * Scale
--
--              The sense terminals ps1 and ns1 behave as a short circuit.
--              The sense terminals ps2 and ns2 behave as a short circuit.
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--  ps1         positive sense terminal for control current 1
--  ns1         negative sense terminal for control current 1
--  ps2         positive sense terminal for control current 2
--  ns2         negative sense terminal for control current 2
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 bufctl vss ctlmain ctlmaini boost boosti IBIS_CCVS_MIN
--
--  VAMS:       IBIS_CCVS_MIN MIN1 (bufctl, vss, ctlmain, vss, boost,
--              vss);
--
-- =========================================================
  entity IBIS_CCVS_MIN is

    generic ( 
      Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_CCVS_MIN;
  --===========================================================================
  architecture ideal of IBIS_CCVS_MIN is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across Iin1 through ps1 to ns1;
  quantity Vin2 across Iin2 through ps2 to ns2;

  begin

    Vin1 == 0.0;
    Vin2 == 0.0;
    Vout == Scale * realmin(Iin1, Iin2);

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_VCVS_MAX (p, n, ps1, ns1, ps2, ns2)
--
-- DESCRIPTION:
--              A voltage-controlled voltage source for taking the greater
--              of two control input voltages. The output voltage value is
--              calculated:
--
--              V(p,n) = min( V(ps1,ns1) V(ps2,ns2) ) * Scale
--
--              The sense terminals ps1 and ns1 draw no current. The sense
--              terminals ps2 and ns2 draw no current.
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--  ps1         positive sense terminal for control voltage 1
--  ns1         negative sense terminal for control voltage 1
--  ps2         positive sense terminal for control voltage 2
--  ns2         negative sense terminal for control voltage 2
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 bufctl vss ctlmain vss boost vss IBIS_VCVS_MAX
--
--  VAMS:       IBIS_VCVS_MAX MAX1 (bufctl, vss, ctlmain, vss, boost,
--              vss);
--
-- =========================================================
  entity IBIS_VCVS_MAX is

    generic ( 
      Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_VCVS_MAX;
  --===========================================================================
  architecture ideal of IBIS_VCVS_MAX is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across              ps1 to ns1;
  quantity Vin2 across              ps2 to ns2;

  begin

    Vout == Scale * realmax(Vin1, Vin2);

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_CCVS_MAX (p, n, ps1, ns1, ps2, ns2)
--
-- DESCRIPTION:
--              A current-controlled voltage source for taking the greater
--              of two input currents. The output voltage value is
--              calculated:
--
--              V(p,n) = max( I(ps1,ns1) I(ps2,ns2) ) * Scale
--
--              The sense terminals ps1 and ns1 draw no current. The sense
--              terminals ps2 and ns2 draw no current.
--
-- TERMINALS:
--
--  p           positive voltage source terminal
--  n           negative voltage source terminal
--  ps1         positive sense terminal for control current 1
--  ns1         negative sense terminal for control current 1
--  ps2         positive sense terminal for control current 2
--  ns2         negative sense terminal for control current 2
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
-- =========================================================
  entity IBIS_CCVS_MAX is

    generic ( 
      Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_CCVS_MAX;
  --===========================================================================
  architecture ideal of IBIS_CCVS_MAX is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across Iin1 through ps1 to ns1;
  quantity Vin2 across Iin2 through ps2 to ns2;

  begin

    Vin1 == 0.0;
    Vin2 == 0.0;
    Vout == Scale * realmax(Iin1, Iin2);

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_VCVS_ABS (p, n, ps, ns)
--
-- DESCRIPTION:
--              A voltage-controlled voltage source for taking the
--              absolute value of the control voltage. The output voltage
--              value is always positive, and is calculated:
--
--              V(p,n) = abs( V(ps,ns) ) * Scale
--
--              The sense terminals ps and ns draw no current.
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--  ps          positive sense terminal for control voltage
--  ns          negative sense terminal for control voltage
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
-- =========================================================
  entity IBIS_VCVS_ABS is

    generic ( 
      Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_VCVS_ABS;
  --===========================================================================
  architecture ideal of IBIS_VCVS_ABS is

  quantity Vout across Iout through p  to n;
  quantity Vin  across              ps to ns;

  begin

    Vout == Scale * abs(Vin);

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_CCVS_ABS (p, n, ps, ns)
--
-- DESCRIPTION:
--              A current-controlled voltage source for taking the
--              absolute value of the control current. The output voltage
--              value is always positive, and is calculated:
--
--              V(p,n) = abs( I(ps,ns) ) * Scale
--
--              The sense terminals ps and ns behave as a short circuit.
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--  ps          positive sense terminal for control current
--  ns          negative sense terminal for control current
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
-- =========================================================
  entity IBIS_CCVS_ABS is

    generic ( 
      Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_CCVS_ABS;
  --===========================================================================
  architecture ideal of IBIS_CCVS_ABS is

  quantity Vout across Iout through p  to n;
  quantity Vin  across Iin  through ps to ns;

  begin

    Vin  == 0.0;
    Vout == Scale * abs(Iin);

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_VCVS_SUM (p, n, ps1, ns1, ps2, ns2)
--
-- DESCRIPTION:
--              A voltage-controlled voltage source for taking the sum of
--              two control input voltages. The output voltage value is
--              calculated:
--
--              V(p,n) = ( V(ps1,ns1) + V(ps2,ns2) ) * Scale
--
--              The sense terminals ps1 and ns1 draw no current. The sense
--              terminals ps2 and ns2 draw no current.
--
-- TERMINALS:
--
--  p           positive voltage source terminal
--  n           negative voltage source terminal
--  ps1         positive sense terminal for control voltage 1
--  ns1         negative sense terminal for control voltage 1
--  ps2         positive sense terminal for control voltage 2
--  ns2         negative sense terminal for control voltage 2
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
-- =========================================================
  entity IBIS_VCVS_SUM is

    generic ( 
      Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_VCVS_SUM;
  --===========================================================================
  architecture ideal of IBIS_VCVS_SUM is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across              ps1 to ns1;
  quantity Vin2 across              ps2 to ns2;

  begin

    Vout == Scale * (Vin1 + Vin2);

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_CCVS_SUM (p, n, ps1, ns1, ps2, ns2)
--
-- DESCRIPTION:
--              A current-controlled voltage source for taking the sum of
--              two control input voltages. The output voltage value is
--              calculated:
--
--              V(p,n) = ( I(ps1,ns1) + I(ps2,ns2) ) * Scale
--
--              The sense terminals ps1 and ns1 behave as a short circuit.
--              The sense terminals ps2 and ns2 behave as a short circuit.
--
-- TERMINALS:
--
--  p           positive voltage source terminal
--  n           negative voltage source terminal
--  ps1         positive sense terminal for control current 1
--  ns1         negative sense terminal for control current 1
--  ps2         positive sense terminal for control current 2
--  ns2         negative sense terminal for control current 2
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
-- =========================================================
  entity IBIS_CCVS_SUM is

    generic ( 
      Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_CCVS_SUM;
  --===========================================================================
  architecture ideal of IBIS_CCVS_SUM is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across Iin1 through ps1 to ns1;
  quantity Vin2 across Iin2 through ps2 to ns2;

  begin

    Vin1 == 0.0;
    Vin2 == 0.0;
    Vout == Scale * (Iin1 + Iin2);

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_VCVS_MULT (p, n, ps1, ns1, ps2, ns2)
--
-- DESCRIPTION:
--              A voltage-controlled voltage source for taking the product
--              of two control input voltages. The output voltage value is
--              calculated:
--
--              V(p,n) = ( V(ps1,ns1) * V(ps2,ns2) ) * Scale
--
--              The sense terminals ps1 and ns1 draw no current. The sense
--              terminals ps2 and ns2 draw no current.
--
-- TERMINALS:
--
--  p           positive voltage source terminal
--  n           negative voltage source terminal
--  ps1         positive sense terminal for control voltage 1
--  ns1         negative sense terminal for control voltage 1
--  ps2         positive sense terminal for control voltage 2
--  ns2         negative sense terminal for control voltage 2
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
-- =========================================================
  entity IBIS_VCVS_MULT is

    generic ( 
      Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_VCVS_MULT;
  --===========================================================================
  architecture ideal of IBIS_VCVS_MULT is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across              ps1 to ns1;
  quantity Vin2 across              ps2 to ns2;

  begin

    Vout == Scale * Vin1 * Vin2;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_CCVS_MULT (p, n, ps1, ns1, ps2, ns2)
--
-- DESCRIPTION:
--              A current-controlled voltage source for taking the product
--              of two control input currents. The output voltage value is
--              calculated:
--
--              V(p,n) = ( I(ps1,ns1) * I(ps2,ns2) ) * Scale
--
--              The sense terminals ps1 and ns1 behave as a short circuit.
--              The sense terminals ps2 and ns2 behave as a short circuit.
--
-- TERMINALS:
--
--  p           positive voltage source terminal
--  n           negative voltage source terminal
--  ps1         positive sense terminal for control current 1
--  ns1         negative sense terminal for control current 1
--  ps2         positive sense terminal for control current 2
--  ns2         negative sense terminal for control current 2
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
-- =========================================================
  entity IBIS_CCVS_MULT is

    generic ( 
      Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_CCVS_MULT;
  --===========================================================================
  architecture ideal of IBIS_CCVS_MULT is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across Iin1 through ps1 to ns1;
  quantity Vin2 across Iin2 through ps2 to ns2;

  begin

    Vin1 == 0.0;
    Vin2 == 0.0;
    Vout == Scale * Iin1 * Iin2;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_VCVS_DIV (p, n, ps1, ns1, ps2, ns2)
--
-- DESCRIPTION:
--              A voltage-controlled voltage source for taking the ratio
--              of two control input voltages. The output voltage value is
--              calculated:
--
--              V(p,n) = ( V(ps1,ns1) / V(ps2,ns2) ) * Scale
--
--              The sense terminals ps1 and ns1 draw no current. The sense
--              terminals ps2 and ns2 draw no current.
--
-- TERMINALS:
--
--  p           positive voltage source terminal
--  n           negative voltage source terminal
--  ps1         positive sense terminal for control voltage 1
--  ns1         negative sense terminal for control voltage 1
--  ps2         positive sense terminal for control voltage 2
--  ns2         negative sense terminal for control voltage 2
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
-- =========================================================
  entity IBIS_VCVS_DIV is

    generic ( 
      Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_VCVS_DIV;
  --===========================================================================
  architecture ideal of IBIS_VCVS_DIV is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across              ps1 to ns1;
  quantity Vin2 across              ps2 to ns2;

  begin

    if Vin2 /= 0.0 use
      Vout == Scale * Vin1 / Vin2;
    else
      Vout == real'high;     -- instead of inf;
    end use;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_CCVS_DIV (p, n, ps1, ns1, ps2, ns2)
--
-- DESCRIPTION:
--              A current-controlled voltage source for taking the ratio
--              of two control input currents. The output voltage value is
--              calculated:
--
--              V(p,n) = ( I(ps1,ns1) / I(ps2,ns2) ) * Scale
--
--              The sense terminals ps1 and ns1 behave as a short circuit.
--              The sense terminals ps2 and ns2 behave as a short circuit.
--
-- TERMINALS:
--
--  p           positive voltage source terminal
--  n           negative voltage source terminal
--  ps1         positive sense terminal for control current 1
--  ns1         negative sense terminal for control current 1
--  ps2         positive sense terminal for control current 2
--  ns2         negative sense terminal for control current 2
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
-- =========================================================
  entity IBIS_CCVS_DIV is

    generic ( 
      Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_CCVS_DIV;
  --===========================================================================
  architecture ideal of IBIS_CCVS_DIV is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across Iin1 through ps1 to ns1;
  quantity Vin2 across Iin2 through ps2 to ns2;

  begin

    Vin1 == 0.0;
    Vin2 == 0.0;
    if Iin2 /= 0.0 use
      Vout == Scale * Iin1 / Iin2;
    else
      Vout == real'high;     -- instead of inf;
    end use;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_VCVS_PWL (p, n, ps, ns)
--
-- DESCRIPTION:
--              A voltage-controlled voltage source with non-linear
--              relationship between control voltage and output voltage,
--              described using a piecewise-linear table.
--
--              Each voltage value in the array of X values is paired with
--              a Y array voltage value. The Length parameter must be set
--              to the number of elements in these arrays.
--
--              The input control voltage determines a position in the X
--              array, and the corersponding Y array value is used for the
--              output voltage. Linear interpolation is used for control
--              voltage values that do not fall exactly on a particular X
--              array value. Points outside the X array range are
--              extrapolated using the slope of the nearest two points of
--              the PWL data.
--
--              The sense terminals ps and ns draw no current.
--
-- TERMINALS:
--
--  p           positive voltage source terminal
--  n           negative voltage source terminal
--  ps          positive sense terminal for control voltage
--  ns          negative sense terminal for control voltage
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--  Length      Number of elements in the X and Y arrays
--              DEFAULT Length = '4'
--
--  X[1:Length]  Array of X axis values (input control voltage)
--              DEFAULT X[1:Length] = '{-5.00 0.00 5.00 10.00}'
--
--  Y[1:Length]  Array of Y axis values (output voltage)
--              DEFAULT Y[1:Length] = '{-0.10 0.00 0.10 0.10}'
--
--
-- INSTANCE CALL EXAMPLES:
--
-- =========================================================
  entity IBIS_VCVS_PWL is

    generic ( 
      Scale  : real    := 1.0;
      -------------------------------------------------------------------------
      -- Vectors of the PWL table
      -------------------------------------------------------------------------
      X : real_vector := (-5.00,  0.00,  5.00, 10.00);
      Y : real_vector := (-0.10,  0.00,  0.10,  0.10));

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_VCVS_PWL;
  --===========================================================================
  architecture ideal of IBIS_VCVS_PWL is

  quantity Vout across Iout through p  to n;
  quantity Vin  across              ps to ns;
  --===========================================================================
  function Lookup (X             : in real;
                   Xdata         : in real_vector;
                   Ydata         : in real_vector;
                   Extrapolation : in string := "SE") return real is 
  -----------------------------------------------------------------------------
  -- This function returns "Y" that corresponds to "X" in the "Ydata" "Xdata"
  -- input pair using linear interpolation.
  --
  -- If the "X" input value lies outside the range of "Xdata", the returned
  -- "Y" value will either be equal to the first or last point in "Ydata",
  -- or it will be calculated using the slope between the first or last two
  -- points of "Ydata".  The extrapolation method is determined by the string
  -- in "Extrapolation".  "HE" stands for "Horizontal Extrapolation" which
  -- selects the former method, and "SE" stands for "Slope Extrapolation" which
  -- selects the latter method.
  --
  -- Acknowledgements:  The original version of this function was developed by
  -- Mentor Graphics Corporation.  Modifications added by Intel Corporation.
  -----------------------------------------------------------------------------
    variable xvalue, yvalue, m : real;
    variable start, fin, mid   : integer; 
  -----------------------------------------------------------------------------
  begin
    ---------------------------------------------------------------------------
    -- Handle cases when "X" is outside the range of "Xdata"
    ---------------------------------------------------------------------------
    if (Extrapolation = "SE") and (X <= Xdata(0)) then
      m := (Ydata(1) - Ydata(0)) / (Xdata(1) - Xdata(0));
      yvalue := Ydata(0) + m * (X - Xdata(0));
      return yvalue;

    elsif (Extrapolation = "HE") and (X <= Xdata(0)) then
      yvalue := Ydata(0);
      return yvalue;

    elsif (Extrapolation = "SE") and (X >= Xdata(Xdata'right)) then
      m := (Ydata(Ydata'right) - Ydata(Ydata'right - 1)) / (Xdata(Xdata'right) - Xdata(Xdata'right - 1));
      yvalue := Ydata(Ydata'right) + m * (X - Xdata(Xdata'right));
      return yvalue;

    elsif (Extrapolation = "HE") and (X >= Xdata(Xdata'right)) then
      yvalue := Ydata(Ydata'right);
      return yvalue;
    ---------------------------------------------------------------------------
    -- Handle cases when "X" is in the range of "Xdata"
    ---------------------------------------------------------------------------
    else
      start:= 0;
      fin := Xdata'right;

      while  start <= fin  loop
        mid := (start + fin) / 2; 

        if Xdata(mid) < X then
          start := mid + 1;
        else fin := mid - 1;
        end if;  

      end loop; 
                       
      if Xdata(mid) > X then
        mid := mid - 1; 
      end if;
      -------------------------------------------------------------------------
      -- Find "Y" by linear interpolation
      -------------------------------------------------------------------------
      yvalue := Ydata(mid) + (X - Xdata(mid)) * (Ydata(mid+1) - Ydata(mid)) / (Xdata(mid+1) - Xdata(mid));
      return yvalue;
    ---------------------------------------------------------------------------
    end if;
    ---------------------------------------------------------------------------
  end function Lookup;
  --===========================================================================

  begin

    Vout == Scale * Lookup(Vin, X, Y, "SE");

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_CCVS_PWL (p, n, ps, ns)
--
-- DESCRIPTION:
--              A current-controlled voltage source with non-linear
--              relationship between control current and output voltage,
--              described using a piecewise-linear table.
--
--              Each current value in the array of X values is paired with
--              a Y array voltage value. The Length parameter must be set
--              to the number of elements in these arrays.
--
--              The input control current determines a position in the X
--              array, and the corersponding Y array value is used for the
--              output voltage. Linear interpolation is used for control
--              current values that do not fall exactly on a particular X
--              array value. Points outside the X array range are
--              extrapolated using the slope of the nearest two points of
--              the PWL data.
--
--              The sense terminals ps and ns behave as a short circuit.
--
-- TERMINALS:
--
--  p           positive voltage source terminal
--  n           negative voltage source terminal
--  ps          positive sense terminal for control current
--  ns          negative sense terminal for control current
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--  Length      Number of elements in the X and Y arrays
--              DEFAULT Length = '4'
--
--  X[1:Length]  Array of X axis values (input control current)
--              DEFAULT X[1:Length] = '{-5.00 0.00 5.00 10.00}'
--
--  Y[1:Length]  Array of Y axis values (output voltage)
--              DEFAULT Y[1:Length] = '{-0.10 0.00 0.10 0.10}'
--
--
-- INSTANCE CALL EXAMPLES:
--
-- =========================================================
  entity IBIS_CCVS_PWL is

    generic ( 
      Scale  : real    := 1.0;
      -------------------------------------------------------------------------
      -- Vectors of the PWL table
      -------------------------------------------------------------------------
      X : real_vector := (-5.00,  0.00,  5.00, 10.00);
      Y : real_vector := (-0.10,  0.00,  0.10,  0.10));

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_CCVS_PWL;
  --===========================================================================
  architecture ideal of IBIS_CCVS_PWL is

  quantity Vout across Iout through p  to n;
  quantity Vin  across Iin  through ps to ns;
  --===========================================================================
  function Lookup (X             : in real;
                   Xdata         : in real_vector;
                   Ydata         : in real_vector;
                   Extrapolation : in string := "SE") return real is 
  -----------------------------------------------------------------------------
  -- This function returns "Y" that corresponds to "X" in the "Ydata" "Xdata"
  -- input pair using linear interpolation.
  --
  -- If the "X" input value lies outside the range of "Xdata", the returned
  -- "Y" value will either be equal to the first or last point in "Ydata",
  -- or it will be calculated using the slope between the first or last two
  -- points of "Ydata".  The extrapolation method is determined by the string
  -- in "Extrapolation".  "HE" stands for "Horizontal Extrapolation" which
  -- selects the former method, and "SE" stands for "Slope Extrapolation" which
  -- selects the latter method.
  --
  -- Acknowledgements:  The original version of this function was developed by
  -- Mentor Graphics Corporation.  Modifications added by Intel Corporation.
  -----------------------------------------------------------------------------
    variable xvalue, yvalue, m : real;
    variable start, fin, mid   : integer; 
  -----------------------------------------------------------------------------
  begin
    ---------------------------------------------------------------------------
    -- Handle cases when "X" is outside the range of "Xdata"
    ---------------------------------------------------------------------------
    if (Extrapolation = "SE") and (X <= Xdata(0)) then
      m := (Ydata(1) - Ydata(0)) / (Xdata(1) - Xdata(0));
      yvalue := Ydata(0) + m * (X - Xdata(0));
      return yvalue;

    elsif (Extrapolation = "HE") and (X <= Xdata(0)) then
      yvalue := Ydata(0);
      return yvalue;

    elsif (Extrapolation = "SE") and (X >= Xdata(Xdata'right)) then
      m := (Ydata(Ydata'right) - Ydata(Ydata'right - 1)) / (Xdata(Xdata'right) - Xdata(Xdata'right - 1));
      yvalue := Ydata(Ydata'right) + m * (X - Xdata(Xdata'right));
      return yvalue;

    elsif (Extrapolation = "HE") and (X >= Xdata(Xdata'right)) then
      yvalue := Ydata(Ydata'right);
      return yvalue;
    ---------------------------------------------------------------------------
    -- Handle cases when "X" is in the range of "Xdata"
    ---------------------------------------------------------------------------
    else
      start:= 0;
      fin := Xdata'right;

      while  start <= fin  loop
        mid := (start + fin) / 2; 

        if Xdata(mid) < X then
          start := mid + 1;
        else fin := mid - 1;
        end if;  

      end loop; 
                       
      if Xdata(mid) > X then
        mid := mid - 1; 
      end if;
      -------------------------------------------------------------------------
      -- Find "Y" by linear interpolation
      -------------------------------------------------------------------------
      yvalue := Ydata(mid) + (X - Xdata(mid)) * (Ydata(mid+1) - Ydata(mid)) / (Xdata(mid+1) - Xdata(mid));
      return yvalue;
    ---------------------------------------------------------------------------
    end if;
    ---------------------------------------------------------------------------
  end function Lookup;
  --===========================================================================

  begin

    Vin  == 0.0;
    Vout == Scale * Lookup(Iin, X, Y, "SE");

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_TCVS_PWL (p, n)
--
-- DESCRIPTION:
--              A time-controlled voltage source with non-linear
--              relationship between time and output voltage, described
--              using a piecewise-linear table.
--
--              Each time value in the array of X values is paired with a
--              Y array votage value. The Length parameter must be set to
--              the number of elements in these arrays.
--
--              The arrays describe the output voltage waveform. The first
--              X value in the array gives the time at wich the waveform
--              begins in the simulation. Linear interpolation is used for
--              time values that do not fall exactly on a particular X
--              array value. Points outside the time range are
--              extrapolated horizontally, ie., the Y value of nearest
--              point is used.
--
-- TERMINALS:
--
--  p           positive voltage source terminal
--  n           negative voltage source terminal
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--  Length      Number of elements in the X and Y arrays
--              DEFAULT Length = '4'
--
--  X[1:Length]  Array of X axis values (time)
--              DEFAULT X[1:Length] = '{0.50e-9 1.00e-9 2.00e-9 2.50e-9}'
--
--  Y[1:Length]  Array of Y axis values (output voltage)
--              DEFAULT Y[1:Length] = '{0.00 0.10 0.90 1.00}'
--
--
-- INSTANCE CALL EXAMPLES:
--
-- =========================================================
  entity IBIS_TCVS_PWL is

    generic ( 
      Scale : real := 1.0;
      -------------------------------------------------------------------------
      -- Vectors of the PWL table
      -------------------------------------------------------------------------
      X : real_vector := (0.50e-9,  1.00e-9,  2.00e-9,  2.50e-9);
      Y : real_vector := (0.00,     0.10,     0.90,     1.00));

    port (terminal p, n : electrical);

  end entity IBIS_TCVS_PWL;
  --===========================================================================
  architecture ideal of IBIS_TCVS_PWL is

  quantity Vout across Iout through p to n;
  --===========================================================================
  function Lookup (X             : in real;
                   Xdata         : in real_vector;
                   Ydata         : in real_vector;
                   Extrapolation : in string := "SE") return real is 
  -----------------------------------------------------------------------------
  -- This function returns "Y" that corresponds to "X" in the "Ydata" "Xdata"
  -- input pair using linear interpolation.
  --
  -- If the "X" input value lies outside the range of "Xdata", the returned
  -- "Y" value will either be equal to the first or last point in "Ydata",
  -- or it will be calculated using the slope between the first or last two
  -- points of "Ydata".  The extrapolation method is determined by the string
  -- in "Extrapolation".  "HE" stands for "Horizontal Extrapolation" which
  -- selects the former method, and "SE" stands for "Slope Extrapolation" which
  -- selects the latter method.
  --
  -- Acknowledgements:  The original version of this function was developed by
  -- Mentor Graphics Corporation.  Modifications added by Intel Corporation.
  -----------------------------------------------------------------------------
    variable xvalue, yvalue, m : real;
    variable start, fin, mid   : integer; 
  -----------------------------------------------------------------------------
  begin
    ---------------------------------------------------------------------------
    -- Handle cases when "X" is outside the range of "Xdata"
    ---------------------------------------------------------------------------
    if (Extrapolation = "SE") and (X <= Xdata(0)) then
      m := (Ydata(1) - Ydata(0)) / (Xdata(1) - Xdata(0));
      yvalue := Ydata(0) + m * (X - Xdata(0));
      return yvalue;

    elsif (Extrapolation = "HE") and (X <= Xdata(0)) then
      yvalue := Ydata(0);
      return yvalue;

    elsif (Extrapolation = "SE") and (X >= Xdata(Xdata'right)) then
      m := (Ydata(Ydata'right) - Ydata(Ydata'right - 1)) / (Xdata(Xdata'right) - Xdata(Xdata'right - 1));
      yvalue := Ydata(Ydata'right) + m * (X - Xdata(Xdata'right));
      return yvalue;

    elsif (Extrapolation = "HE") and (X >= Xdata(Xdata'right)) then
      yvalue := Ydata(Ydata'right);
      return yvalue;
    ---------------------------------------------------------------------------
    -- Handle cases when "X" is in the range of "Xdata"
    ---------------------------------------------------------------------------
    else
      start:= 0;
      fin := Xdata'right;

      while  start <= fin  loop
        mid := (start + fin) / 2; 

        if Xdata(mid) < X then
          start := mid + 1;
        else fin := mid - 1;
        end if;  

      end loop; 
                       
      if Xdata(mid) > X then
        mid := mid - 1; 
      end if;
      -------------------------------------------------------------------------
      -- Find "Y" by linear interpolation
      -------------------------------------------------------------------------
      yvalue := Ydata(mid) + (X - Xdata(mid)) * (Ydata(mid+1) - Ydata(mid)) / (Xdata(mid+1) - Xdata(mid));
      return yvalue;
    ---------------------------------------------------------------------------
    end if;
    ---------------------------------------------------------------------------
  end function Lookup;
  --===========================================================================

  begin

    Vout == Scale * Lookup(now, X, Y, "HE");

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_VECVS_PWL (p, n, ps, ns)
--
-- DESCRIPTION:
--              A voltage event-controlled voltage source producing
--              time-based piecewise-linear voltage waveforms in response
--              to trigger events. The control voltage triggers separate
--              rising and falling PWL waveforms to effect transitions
--              between high and low states.
--
--              Each time value in the XR array is paired with a voltage
--              value in the YR array to describe a rising edge waveform.
--              Likewise for XF and YF arrays, which describe a falling
--              edge waveform. The Length_R and Length_F parameters must
--              be set to the number of elements in the *R and *F arrays,
--              respectively.
--
--              When the control voltage rises above Vth_R the rising edge
--              waveform is output. Voltage then remains at the level of
--              the last point in the rising waveform table until a
--              falling edge transition is triggerd. When the control
--              voltage falls below Vth_F a falling edge waveform is
--              triggered, holding at the last voltage until the next
--              rising edge.
--
--              In DC analysis the first point of the falling edge
--              waveform is output if the control voltage is above Vth_R,
--              and the first point of the rising edge waveform is output
--              if the control voltage is below Vth_F. If Vt_R is less
--              than or equal to Vth_F and the control voltage lies on or
--              between those values the high state voltage is output. If
--              Vth_R is greater than Vth_F and the control voltage lies
--              on or between those values, the average of the high and
--              low state voltages is output.
--
-- TERMINALS:
--
--  p           positive voltage source terminal
--  n           negative voltage source terminal
--  ps          positive sense terminal for control voltage
--  ns          negative sense terminal for control voltage
--
-- PARAMETERS:
--
--  Vth_R       Threshold voltage to trigger a rising edge event
--              DEFAULT Vth_R = '0.0'
--
--  Vth_F       Threshold voltage to trigger a falling edge event
--              DEFAULT Vth_F = '0.0'
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--  Length_R    Number of elements in the XR and YR arrays
--              DEFAULT Length_R = '4'
--
--  Length_F    Number of elements in the XF and YF arrays
--              DEFAULT Length_F = '4'
--
--  XR[1:Length_R]  Array of rising edge X values (time)
--              DEFAULT XR[1:Length_R] = '{0.50e-9 1.00e-9 2.00e-9 2.50e-9}'
--
--  YR[1:Length_R]  Array of rising edge Y values (voltage)
--              DEFAULT YR[1:Length_R] = '{0.00 0.10 0.90 1.00}'
--
--  XF[1:Length_F]  Array of falling edge X values (time)
--              DEFAULT XF[1:Length_F] = '{0.50e-9 1.00e-9 2.00e-9 2.50e-9}'
--
--  YF[1:Length_F]  Array of falling edge Y values (voltage)
--              DEFAULT YF[1:Length_F] = '{1.00 0.90 0.10 0.00}'
--
--
-- INSTANCE CALL EXAMPLES:
--
-- =========================================================
  entity IBIS_VECVS_PWL is

    generic ( 
      Vth_R  : real    := 0.0;
      Vth_F  : real    := 0.0;
      Scale  : real    := 1.0;
      -------------------------------------------------------------------------
      -- Vectors of the PWL table
      -------------------------------------------------------------------------
      XR : real_vector := (0.50e-9,  1.00e-9,  2.00e-9,  2.50e-9);
      YR : real_vector := (0.00,     0.10,     0.90,     1.00);
      XF : real_vector := (0.50e-9,  1.00e-9,  2.00e-9,  2.50e-9);
      YF : real_vector := (1.00,     0.90,     0.10,     0.00));

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_VECVS_PWL;
  --===========================================================================
  architecture ideal of IBIS_VECVS_PWL is

  quantity Vout across Iout through p  to n;
  quantity Vin  across              ps to ns;
  signal   T_event    : real    := 0.0;
  signal   EventState : integer := 0;
  --===========================================================================
  function Lookup (X             : in real;
                   Xdata         : in real_vector;
                   Ydata         : in real_vector;
                   Extrapolation : in string := "SE") return real is 
  -----------------------------------------------------------------------------
  -- This function returns "Y" that corresponds to "X" in the "Ydata" "Xdata"
  -- input pair using linear interpolation.
  --
  -- If the "X" input value lies outside the range of "Xdata", the returned
  -- "Y" value will either be equal to the first or last point in "Ydata",
  -- or it will be calculated using the slope between the first or last two
  -- points of "Ydata".  The extrapolation method is determined by the string
  -- in "Extrapolation".  "HE" stands for "Horizontal Extrapolation" which
  -- selects the former method, and "SE" stands for "Slope Extrapolation" which
  -- selects the latter method.
  --
  -- Acknowledgements:  The original version of this function was developed by
  -- Mentor Graphics Corporation.  Modifications added by Intel Corporation.
  -----------------------------------------------------------------------------
    variable xvalue, yvalue, m : real;
    variable start, fin, mid   : integer; 
  -----------------------------------------------------------------------------
  begin
    ---------------------------------------------------------------------------
    -- Handle cases when "X" is outside the range of "Xdata"
    ---------------------------------------------------------------------------
    if (Extrapolation = "SE") and (X <= Xdata(0)) then
      m := (Ydata(1) - Ydata(0)) / (Xdata(1) - Xdata(0));
      yvalue := Ydata(0) + m * (X - Xdata(0));
      return yvalue;

    elsif (Extrapolation = "HE") and (X <= Xdata(0)) then
      yvalue := Ydata(0);
      return yvalue;

    elsif (Extrapolation = "SE") and (X >= Xdata(Xdata'right)) then
      m := (Ydata(Ydata'right) - Ydata(Ydata'right - 1)) / (Xdata(Xdata'right) - Xdata(Xdata'right - 1));
      yvalue := Ydata(Ydata'right) + m * (X - Xdata(Xdata'right));
      return yvalue;

    elsif (Extrapolation = "HE") and (X >= Xdata(Xdata'right)) then
      yvalue := Ydata(Ydata'right);
      return yvalue;
    ---------------------------------------------------------------------------
    -- Handle cases when "X" is in the range of "Xdata"
    ---------------------------------------------------------------------------
    else
      start:= 0;
      fin := Xdata'right;

      while  start <= fin  loop
        mid := (start + fin) / 2; 

        if Xdata(mid) < X then
          start := mid + 1;
        else fin := mid - 1;
        end if;  

      end loop; 
                       
      if Xdata(mid) > X then
        mid := mid - 1; 
      end if;
      -------------------------------------------------------------------------
      -- Find "Y" by linear interpolation
      -------------------------------------------------------------------------
      yvalue := Ydata(mid) + (X - Xdata(mid)) * (Ydata(mid+1) - Ydata(mid)) / (Xdata(mid+1) - Xdata(mid));
      return yvalue;
    ---------------------------------------------------------------------------
    end if;
    ---------------------------------------------------------------------------
  end function Lookup;
  --===========================================================================

  begin

    ---------------------------------------------------------------------------
    Threshold_crossing : process is
    begin
      if domain = quiescent_domain then
        wait until domain /= quiescent_domain;
        if Vin > Vth_R then            -- high state
          EventState <= 1;
        elsif Vin < Vth_F then         -- low state
          EventState <= -1;
        end if;
      else
        wait on Vin'above(Vth_R), Vin'above(Vth_F);
        if Vin > Vth_R then            -- rising edge
          EventState <= 2;
          T_event    <= now;
        elsif Vin < Vth_F then         -- falling edge
          EventState <= -2;
          T_event    <= now;
        end if;
      end if;
    end process Threshold_crossing;
    ---------------------------------------------------------------------------

    case EventState use
      when -2 =>        -- Executes after falling edge events
        Vout == Scale * Lookup(now-T_event, XF, YF, "HE");
      when -1 =>        -- Executes between t=0 and first threshold crossing
        Vout == Scale * Lookup(0.0, XR, YR, "HE");    -- when input is LOW
      when 1 =>         -- Executes between t=0 and first threshold crossing
        Vout == Scale * Lookup(0.0, XF, YF, "HE");    -- when input is HIGH
      when 2 =>         -- Executes after rising edge events
        Vout == Scale * Lookup(now-T_event, XR, YR, "HE");
      when others =>    -- Executes during DC operating point analysis
        if Vin > Vth_R use
          Vout == Scale * Lookup(0.0, XF, YF, "HE");
        elsif Vin < Vth_F use
          Vout == Scale * Lookup(0.0, XR, YR, "HE");
        else
          Vout == Scale * (Lookup(0.0, XR, YR, "HE") + Lookup(0.0, XF, YF, "HE")) / 2.0;
        end use;
    end case;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_CECVS_PWL (p, n, ps, ns)
--
-- DESCRIPTION:
--              A current event-controlled voltage source producing
--              time-based piecewise-linear voltage waveforms in response
--              to trigger events. The control current triggers separate
--              rising and falling PWL waveforms to effect transitions
--              between high and low states.
--
--              Each time value in the XR array is paired with a voltage
--              value in the YR array to describe a rising edge waveform.
--              Likewise for XF and YF arrays, which describe a falling
--              edge waveform. The Length_R and Length_F parameters must
--              be set to the number of elements in the *R and *F arrays,
--              respectively.
--
--              When the control current rises above Ith_R the rising edge
--              waveform is output. Voltage then remains at the level of
--              the last point in the rising waveform table until a
--              falling edge transition is triggerd. When the control
--              current falls below Ith_F a falling edge waveform is
--              triggered, holding at the last voltage until the next
--              rising edge.
--
--              In DC analysis the first point of the falling edge
--              waveform is output if the control current is above Ith_R,
--              and the first point of the rising edge waveform is output
--              if the control current is below Ith_F. If It_R is less
--              than or equal to Ith_F and the control current lies on or
--              between those values the high state voltage is output. If
--              Ith_R is greater than Ith_F and the control current lies
--              on or between those values, the average of the high and
--              low state voltages is output.
--
-- TERMINALS:
--
--  p           positive voltage source terminal
--  n           negative voltage source terminal
--  ps          positive sense terminal for control current
--  ns          negative sense terminal for control current
--
-- PARAMETERS:
--
--  Ith_R       Threshold current to trigger a rising edge event
--              DEFAULT Ith_R = '0.0'
--
--  Ith_F       Threshold current to trigger a falling edge event
--              DEFAULT Ith_F = '0.0'
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--  Length_R    Number of elements in the XR and YR arrays
--              DEFAULT Length_R = '4'
--
--  Length_F    Number of elements in the XF and YF arrays
--              DEFAULT Length_F = '4'
--
--  XR[1:Length_R]  Array of rising edge X values (time)
--              DEFAULT XR[1:Length_R] = '{0.50e-9 1.00e-9 2.00e-9 2.50e-9}'
--
--  YR[1:Length_R]  Array of rising edge Y values (current)
--              DEFAULT YR[1:Length_R] = '{0.00 0.10 0.90 1.00}'
--
--  XF[1:Length_F]  Array of falling edge X values (time)
--              DEFAULT XF[1:Length_F] = '{0.50e-9 1.00e-9 2.00e-9 2.50e-9}'
--
--  YF[1:Length_F]  Array of falling edge Y values (current)
--              DEFAULT YF[1:Length_F] = '{1.00 0.90 0.10 0.00}'
--
--
-- INSTANCE CALL EXAMPLES:
--
-- =========================================================
  entity IBIS_CECVS_PWL is

    generic ( 
      Ith_R  : real    := 0.0;
      Ith_F  : real    := 0.0;
      Scale  : real    := 1.0;
      -------------------------------------------------------------------------
      -- Vectors of the PWL table
      -------------------------------------------------------------------------
      XR : real_vector := (0.50e-9,  1.00e-9,  2.00e-9,  2.50e-9);
      YR : real_vector := (0.00,     0.10,     0.90,     1.00);
      XF : real_vector := (0.50e-9,  1.00e-9,  2.00e-9,  2.50e-9);
      YF : real_vector := (1.00,     0.90,     0.10,     0.00));

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_CECVS_PWL;
  --===========================================================================
  architecture ideal of IBIS_CECVS_PWL is

  quantity Vout across Iout through p  to n;
  quantity Vin  across Iin  through ps to ns;
  signal   T_event    : real    := 0.0;
  signal   EventState : integer := 0;
  --===========================================================================
  function Lookup (X             : in real;
                   Xdata         : in real_vector;
                   Ydata         : in real_vector;
                   Extrapolation : in string := "SE") return real is 
  -----------------------------------------------------------------------------
  -- This function returns "Y" that corresponds to "X" in the "Ydata" "Xdata"
  -- input pair using linear interpolation.
  --
  -- If the "X" input value lies outside the range of "Xdata", the returned
  -- "Y" value will either be equal to the first or last point in "Ydata",
  -- or it will be calculated using the slope between the first or last two
  -- points of "Ydata".  The extrapolation method is determined by the string
  -- in "Extrapolation".  "HE" stands for "Horizontal Extrapolation" which
  -- selects the former method, and "SE" stands for "Slope Extrapolation" which
  -- selects the latter method.
  --
  -- Acknowledgements:  The original version of this function was developed by
  -- Mentor Graphics Corporation.  Modifications added by Intel Corporation.
  -----------------------------------------------------------------------------
    variable xvalue, yvalue, m : real;
    variable start, fin, mid   : integer; 
  -----------------------------------------------------------------------------
  begin
    ---------------------------------------------------------------------------
    -- Handle cases when "X" is outside the range of "Xdata"
    ---------------------------------------------------------------------------
    if (Extrapolation = "SE") and (X <= Xdata(0)) then
      m := (Ydata(1) - Ydata(0)) / (Xdata(1) - Xdata(0));
      yvalue := Ydata(0) + m * (X - Xdata(0));
      return yvalue;

    elsif (Extrapolation = "HE") and (X <= Xdata(0)) then
      yvalue := Ydata(0);
      return yvalue;

    elsif (Extrapolation = "SE") and (X >= Xdata(Xdata'right)) then
      m := (Ydata(Ydata'right) - Ydata(Ydata'right - 1)) / (Xdata(Xdata'right) - Xdata(Xdata'right - 1));
      yvalue := Ydata(Ydata'right) + m * (X - Xdata(Xdata'right));
      return yvalue;

    elsif (Extrapolation = "HE") and (X >= Xdata(Xdata'right)) then
      yvalue := Ydata(Ydata'right);
      return yvalue;
    ---------------------------------------------------------------------------
    -- Handle cases when "X" is in the range of "Xdata"
    ---------------------------------------------------------------------------
    else
      start:= 0;
      fin := Xdata'right;

      while  start <= fin  loop
        mid := (start + fin) / 2; 

        if Xdata(mid) < X then
          start := mid + 1;
        else fin := mid - 1;
        end if;  

      end loop; 
                       
      if Xdata(mid) > X then
        mid := mid - 1; 
      end if;
      -------------------------------------------------------------------------
      -- Find "Y" by linear interpolation
      -------------------------------------------------------------------------
      yvalue := Ydata(mid) + (X - Xdata(mid)) * (Ydata(mid+1) - Ydata(mid)) / (Xdata(mid+1) - Xdata(mid));
      return yvalue;
    ---------------------------------------------------------------------------
    end if;
    ---------------------------------------------------------------------------
  end function Lookup;
  --===========================================================================

  begin

    ---------------------------------------------------------------------------
    Threshold_crossing : process is
    begin
      if domain = quiescent_domain then
        wait until domain /= quiescent_domain;
        if Iin > Ith_R then            -- high state
          EventState <= 1;
        elsif Iin < Ith_F then         -- low state
          EventState <= -1;
        end if;
      else
        wait on Iin'above(Ith_R), Iin'above(Ith_F);
        if Iin > Ith_R then            -- rising edge
          EventState <= 2;
          T_event    <= now;
        elsif Iin < Ith_F then         -- falling edge
          EventState <= -2;
          T_event    <= now;
        end if;
      end if;
    end process Threshold_crossing;
    ---------------------------------------------------------------------------

    Vin == 0.0;
    case EventState use
      when -2 =>        -- Executes after falling edge events
        Vout == Scale * Lookup(now-T_event, XF, YF, "HE");
      when -1 =>        -- Executes between t=0 and first threshold crossing
        Vout == Scale * Lookup(0.0, XR, YR, "HE");    -- when input is LOW
      when 1 =>         -- Executes between t=0 and first threshold crossing
        Vout == Scale * Lookup(0.0, XF, YF, "HE");    -- when input is HIGH
      when 2 =>         -- Executes after rising edge events
        Vout == Scale * Lookup(now-T_event, XR, YR, "HE");
      when others =>    -- Executes during DC operating point analysis
        if Iin > Ith_R use
          Vout == Scale * Lookup(0.0, XF, YF, "HE");
        elsif Iin < Ith_F use
          Vout == Scale * Lookup(0.0, XR, YR, "HE");
        else
          Vout == Scale * (Lookup(0.0, XR, YR, "HE") + Lookup(0.0, XF, YF, "HE")) / 2.0;
        end use;
    end case;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_I (p, n)
--
-- DESCRIPTION:
--              A DC current source. Output current is calculated:
--
--              I(p,n) = Idc * Scale
--
-- TERMINALS:
--
--  p           positive current source terminal
--  n           negative current source terminal
--
-- PARAMETERS:
--
--  Idc         DC current value
--              DEFAULT Idc = '1.0'
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 pada vss IBIS_I Idc=0.025
--
--  VAMS:       IBIS_I #(.Idc(0.025)) I1 (pada, vss);
--
-- =========================================================
  entity IBIS_I is

    generic ( 
      Idc   : real := 1.0;
      Scale : real := 1.0);

    port (terminal p, n : electrical);

  end entity IBIS_I;
  --===========================================================================
  architecture ideal of IBIS_I is

  quantity Vout across Iout through p to n;

  begin

    Iout == Scale * Idc;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_VCCS (p, n, ps, ns)
--
-- DESCRIPTION:
--              A voltage-controlled current source. The output current
--              value is calculated:
--
--              I(p,n) = I(ps,ns) * Scale
--
--              The sense terminals ps and ns draw no current.
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--  ps          positive sense terminal for control voltage
--  ns          negative sense terminal for control voltage
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output current
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 pupwr pupwra vcp vss IBIS_VCCS Scale=0.05
--
--  VAMS:       IBIS_VCCS #(.Scale(0.05)) VCCS1 (pupwr, pupwr, vcp, vss);
--
-- =========================================================
  entity IBIS_VCCS is

    generic ( 
      Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_VCCS;
  --===========================================================================
  architecture ideal of IBIS_VCCS is

  quantity Vout across Iout through p  to n;
  quantity Vin  across              ps to ns;

  begin

    Iout == Scale * Vin;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_CCCS (p, n, ps, ns)
--
-- DESCRIPTION:
--              A current-controlled current source. The output current
--              value is calculated:
--
--              I(p,n) = I(ps,ns) * Scale
--
--              The sense terminals ps and ns behave as a short circuit.
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--  ps          positive sense terminal for control current
--  ns          negative sense terminal for control current
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output current
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 buf2in vss pad padi IBIS_CCVS Scale=0.1
--
--  VAMS:       IBIS_CCVS #(.Scale(0.1)) CCVS1 (buf2in, vss, pad, padi);
--
-- =========================================================
  entity IBIS_CCCS is

    generic ( 
      Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_CCCS;
  --===========================================================================
  architecture ideal of IBIS_CCCS is

  quantity Vout across Iout through p  to n;
  quantity Vin  across Iin  through ps to ns;

  begin

    Vin  == 0.0;
    Iout == Scale * Iin;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_VCCS_DELAY (p, n, ps, ns)
--
-- DESCRIPTION:
--              A voltage-controlled current source with delay. All
--              changes at the sense terminals affect the output later in
--              time, using the specified delay. The output current value
--              is calculated:
--
--              I(p,n) = V(ps,ns) * Scale, delayed by TD
--
--              The sense terminals ps and ns draw no current.
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--  ps          positive sense terminal for control voltage
--  ns          negative sense terminal for control voltage
--
-- PARAMETERS:
--
--  TD          Time delay from control voltage change to output change
--              DEFAULT TD = '0.0'
--
--  Scale       Scaling factor for calculating output current
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 buf2in vss pad padi IBIS_VCCS_DELAY TD=200p
--
--  VAMS:       IBIS_VCCS_DELAY #(.TD(200e-12)) \
--              VCCS1 (buf2in, vss, pad, padi);
--
-- =========================================================
  entity IBIS_VCCS_DELAY is

    generic ( 
      TD    : real := 0.0;
      Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_VCCS_DELAY;
  --===========================================================================
  architecture ideal of IBIS_VCCS_DELAY is

  quantity Vout across Iout through p  to n;
  quantity Vin  across              ps to ns;

  begin

    Iout == Scale * Vin'delayed(TD);

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_CCCS_DELAY (p, n, ps, ns)
--
-- DESCRIPTION:
--              A current-controlled current source with delay. All
--              changes at the sense terminals affect the output later,
--              using the specified delay in seconds. The current value is
--              calculated:
--
--              I(p,n) = I(ps,ns) * Scale
--
--              The sense terminals ps and ns behave as a short circuit.
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--  ps          positive sense terminal for control current
--  ns          negative sense terminal for control current
--
-- PARAMETERS:
--
--  TD          Time delay from control current change to output change
--              DEFAULT TD = '0.0'
--
--  Scale       Scaling factor for calculating current
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 buf2in vss pad padi IBIS_CCVS_DELAY Scale=100 TD=200p
--
--  VAMS:       IBIS_CCVS_DELAY #(.Scale(100),.TD(200e-12)) \
--              CCVS1 (buf2in, vss, pad, padi);
--
-- =========================================================
  entity IBIS_CCCS_DELAY is

    generic ( 
      TD    : real := 0.0;
      Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_CCCS_DELAY;
  --===========================================================================
  architecture ideal of IBIS_CCCS_DELAY is

  quantity Vout across Iout through p  to n;
  quantity Vin  across Iin  through ps to ns;
  quantity Iin_as_real  : real;   -- Delete these two lines when the last line
  quantity Iin_real_del : real;   -- in the code works

  begin

    Vin  == 0.0;
    Iin_as_real  == Iin;
    Iin_real_del == Iin_as_real'delayed(TD);
    Iout == Scale * Iin_real_del;
--  Iout == Scale * Iin'delayed(TD);   -- Delete the above 3 lines if this works

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_VCCS_MIN (p, n, ps1, ns1, ps2, ns2)
--
-- DESCRIPTION:
--              A voltage-controlled current source for taking the lesser
--              of two control input voltages. The output current value is
--              calculated:
--
--              I(p,n) = min( V(ps1,ns1) V(ps2,ns2) ) * Scale
--
--              The sense terminals ps1 and ns1 draw no current. The sense
--              terminals ps2 and ns2 draw no current.
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--  ps1         positive sense terminal for control voltage 1
--  ns1         negative sense terminal for control voltage 1
--  ps2         positive sense terminal for control voltage 2
--  ns2         negative sense terminal for control voltage 2
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output current
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 bufctl vss ctlmain vss boost vss IBIS_VCCS_MIN
--
--  VAMS:       IBIS_VCCS_MIN MIN1 (bufctl, vss, ctlmain, vss, boost,
--              vss);
--
-- =========================================================
  entity IBIS_VCCS_MIN is

    generic ( 
      Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_VCCS_MIN;
  --===========================================================================
  architecture ideal of IBIS_VCCS_MIN is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across              ps1 to ns1;
  quantity Vin2 across              ps2 to ns2;

  begin

    Iout == Scale * realmin(Vin1, Vin2);

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_CCCS_MIN (p, n, ps1, ns1, ps2, ns2)
--
-- DESCRIPTION:
--              A voltage-controlled current source for taking the lesser
--              of two control input voltages. The output current value is
--              calculated:
--
--              I(p,n) = min( V(ps1,ns1) V(ps2,ns2) ) * Scale
--
--              The sense terminals ps1 and ns1 behave as a short circuit.
--              The sense terminals ps2 and ns2 behave as a short circuit.
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--  ps1         positive sense terminal for control current 1
--  ns1         negative sense terminal for control current 1
--  ps2         positive sense terminal for control current 2
--  ns2         negative sense terminal for control current 2
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output current
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 bufctl vss ctlmain ctlmaini boost boosti IBIS_CCVS_MIN
--
--  VAMS:       IBIS_CCVS_MIN MIN1 (bufctl, vss, ctlmain, vss, boost,
--              vss);
--
-- =========================================================
  entity IBIS_CCCS_MIN is

    generic ( 
      Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_CCCS_MIN;
  --===========================================================================
  architecture ideal of IBIS_CCCS_MIN is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across Iin1 through ps1 to ns1;
  quantity Vin2 across Iin2 through ps2 to ns2;

  begin

    Vin1 == 0.0;
    Vin2 == 0.0;
    Iout == Scale * realmin(Iin1, Iin2);

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_VCCS_MAX (p, n, ps1, ns1, ps2, ns2)
--
-- DESCRIPTION:
--              A voltage-controlled current source for taking the greater
--              of two control input voltages. The output current value is
--              calculated:
--
--              I(p,n) = min( V(ps1,ns1) V(ps2,ns2) ) * Scale
--
--              The sense terminals ps1 and ns1 draw no current. The sense
--              terminals ps2 and ns2 draw no current.
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--  ps1         positive sense terminal for control voltage 1
--  ns1         negative sense terminal for control voltage 1
--  ps2         positive sense terminal for control voltage 2
--  ns2         negative sense terminal for control voltage 2
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output current
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 bufctl vss ctlmain vss boost vss IBIS_VCCS_MAX
--
--  VAMS:       IBIS_VCCS_MAX MAX1 (bufctl, vss, ctlmain, vss, boost,
--              vss);
--
-- =========================================================
  entity IBIS_VCCS_MAX is

    generic ( 
      Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_VCCS_MAX;
  --===========================================================================
  architecture ideal of IBIS_VCCS_MAX is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across              ps1 to ns1;
  quantity Vin2 across              ps2 to ns2;

  begin

    Iout == Scale * realmax(Vin1, Vin2);

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_CCCS_MAX (p, n, ps1, ns1, ps2, ns2)
--
-- DESCRIPTION:
--              A current-controlled current source for taking the greater
--              of two input currents. The output current value is
--              calculated:
--
--              V(p,n) = max( I(ps1,ns1) I(ps2,ns2) ) * Scale
--
--              The sense terminals ps1 and ns1 draw no current. The sense
--              terminals ps2 and ns2 draw no current.
--
-- TERMINALS:
--
--  p           positive current source terminal
--  n           negative current source terminal
--  ps1         positive sense terminal for control current 1
--  ns1         negative sense terminal for control current 1
--  ps2         positive sense terminal for control current 2
--  ns2         negative sense terminal for control current 2
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output current
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
-- =========================================================
  entity IBIS_CCCS_MAX is

    generic ( 
      Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_CCCS_MAX;
  --===========================================================================
  architecture ideal of IBIS_CCCS_MAX is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across Iin1 through ps1 to ns1;
  quantity Vin2 across Iin2 through ps2 to ns2;

  begin

    Vin1 == 0.0;
    Vin2 == 0.0;
    Iout == Scale * realmax(Iin1, Iin2);

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_VCCS_ABS (p, n, ps, ns)
--
-- DESCRIPTION:
--              A voltage-controlled current source for taking the
--              absolute value of the control voltage. The output current
--              value is always positive, and is calculated:
--
--              V(p,n) = abs( V(ps,ns) ) * Scale
--
--              The sense terminals ps and ns draw no current.
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--  ps          positive sense terminal for control voltage
--  ns          negative sense terminal for control voltage
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output current
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
-- =========================================================
  entity IBIS_VCCS_ABS is

    generic ( 
      Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_VCCS_ABS;
  --===========================================================================
  architecture ideal of IBIS_VCCS_ABS is

  quantity Vout across Iout through p  to n;
  quantity Vin  across              ps to ns;

  begin

    Iout == Scale * abs(Vin);

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_CCCS_ABS (p, n, ps, ns)
--
-- DESCRIPTION:
--              A current-controlled current source for taking the
--              absolute value of the control current. The output current
--              value is always positive, and is calculated:
--
--              V(p,n) = abs( I(ps,ns) ) * Scale
--
--              The sense terminals ps and ns behave as a short circuit.
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--  ps          positive sense terminal for control current
--  ns          negative sense terminal for control current
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output current
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
-- =========================================================
  entity IBIS_CCCS_ABS is

    generic ( 
      Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_CCCS_ABS;
  --===========================================================================
  architecture ideal of IBIS_CCCS_ABS is

  quantity Vout across Iout through p  to n;
  quantity Vin  across Iin  through ps to ns;

  begin

    Vin  == 0.0;
    Iout == Scale * abs(Iin);

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_VCCS_SUM (p, n, ps1, ns1, ps2, ns2)
--
-- DESCRIPTION:
--              A voltage-controlled current source for taking the sum of
--              two control input voltages. The output current value is
--              calculated:
--
--              V(p,n) = ( V(ps1,ns1) + V(ps2,ns2) ) * Scale
--
--              The sense terminals ps1 and ns1 draw no current. The sense
--              terminals ps2 and ns2 draw no current.
--
-- TERMINALS:
--
--  p           positive current source terminal
--  n           negative current source terminal
--  ps1         positive sense terminal for control voltage 1
--  ns1         negative sense terminal for control voltage 1
--  ps2         positive sense terminal for control voltage 2
--  ns2         negative sense terminal for control voltage 2
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output current
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
-- =========================================================
  entity IBIS_VCCS_SUM is

    generic ( 
      Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_VCCS_SUM;
  --===========================================================================
  architecture ideal of IBIS_VCCS_SUM is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across              ps1 to ns1;
  quantity Vin2 across              ps2 to ns2;

  begin

    Iout == Scale * (Vin1 + Vin2);

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_CCCS_SUM (p, n, ps1, ns1, ps2, ns2)
--
-- DESCRIPTION:
--              A current-controlled current source for taking the sum of
--              two control input voltages. The output current value is
--              calculated:
--
--              V(p,n) = ( I(ps1,ns1) + I(ps2,ns2) ) * Scale
--
--              The sense terminals ps1 and ns1 behave as a short circuit.
--              The sense terminals ps2 and ns2 behave as a short circuit.
--
-- TERMINALS:
--
--  p           positive current source terminal
--  n           negative current source terminal
--  ps1         positive sense terminal for control current 1
--  ns1         negative sense terminal for control current 1
--  ps2         positive sense terminal for control current 2
--  ns2         negative sense terminal for control current 2
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output current
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
-- =========================================================
  entity IBIS_CCCS_SUM is

    generic ( 
      Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_CCCS_SUM;
  --===========================================================================
  architecture ideal of IBIS_CCCS_SUM is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across Iin1 through ps1 to ns1;
  quantity Vin2 across Iin2 through ps2 to ns2;

  begin

    Vin1 == 0.0;
    Vin2 == 0.0;
    Iout == Scale * (Iin1 + Iin2);

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_VCCS_MULT (p, n, ps1, ns1, ps2, ns2)
--
-- DESCRIPTION:
--              A voltage-controlled current source for taking the product
--              of two control input voltages. The output current value is
--              calculated:
--
--              V(p,n) = ( V(ps1,ns1) * V(ps2,ns2) ) * Scale
--
--              The sense terminals ps1 and ns1 draw no current. The sense
--              terminals ps2 and ns2 draw no current.
--
-- TERMINALS:
--
--  p           positive current source terminal
--  n           negative current source terminal
--  ps1         positive sense terminal for control voltage 1
--  ns1         negative sense terminal for control voltage 1
--  ps2         positive sense terminal for control voltage 2
--  ns2         negative sense terminal for control voltage 2
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output current
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
-- =========================================================
  entity IBIS_VCCS_MULT is

    generic ( 
      Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_VCCS_MULT;
  --===========================================================================
  architecture ideal of IBIS_VCCS_MULT is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across              ps1 to ns1;
  quantity Vin2 across              ps2 to ns2;

  begin

    Iout == Scale * Vin1 * Vin2;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_CCCS_MULT (p, n, ps1, ns1, ps2, ns2)
--
-- DESCRIPTION:
--              A current-controlled current source for taking the product
--              of two control input currents. The output current value is
--              calculated:
--
--              V(p,n) = ( I(ps1,ns1) * I(ps2,ns2) ) * Scale
--
--              The sense terminals ps1 and ns1 behave as a short circuit.
--              The sense terminals ps2 and ns2 behave as a short circuit.
--
-- TERMINALS:
--
--  p           positive current source terminal
--  n           negative current source terminal
--  ps1         positive sense terminal for control current 1
--  ns1         negative sense terminal for control current 1
--  ps2         positive sense terminal for control current 2
--  ns2         negative sense terminal for control current 2
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output current
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
-- =========================================================
  entity IBIS_CCCS_MULT is

    generic ( 
      Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_CCCS_MULT;
  --===========================================================================
  architecture ideal of IBIS_CCCS_MULT is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across Iin1 through ps1 to ns1;
  quantity Vin2 across Iin2 through ps2 to ns2;

  begin

    Vin1 == 0.0;
    Vin2 == 0.0;
    Iout == Scale * Iin1 * Iin2;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_VCCS_DIV (p, n, ps1, ns1, ps2, ns2)
--
-- DESCRIPTION:
--              A voltage-controlled current source for taking the ratio
--              of two control input voltages. The output current value is
--              calculated:
--
--              V(p,n) = ( V(ps1,ns1) / V(ps2,ns2) ) * Scale
--
--              The sense terminals ps1 and ns1 draw no current. The sense
--              terminals ps2 and ns2 draw no current.
--
-- TERMINALS:
--
--  p           positive current source terminal
--  n           negative current source terminal
--  ps1         positive sense terminal for control voltage 1
--  ns1         negative sense terminal for control voltage 1
--  ps2         positive sense terminal for control voltage 2
--  ns2         negative sense terminal for control voltage 2
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output current
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
-- =========================================================
  entity IBIS_VCCS_DIV is

    generic ( 
      Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_VCCS_DIV;
  --===========================================================================
  architecture ideal of IBIS_VCCS_DIV is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across              ps1 to ns1;
  quantity Vin2 across              ps2 to ns2;

  begin

    if Vin2 /= 0.0 use
      Iout == Scale * Vin1 / Vin2;
    else
      Iout == real'high;     -- instead of inf;
    end use;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_CCCS_DIV (p, n, ps1, ns1, ps2, ns2)
--
-- DESCRIPTION:
--              A current-controlled current source for taking the ratio
--              of two control input currents. The output current value is
--              calculated:
--
--              V(p,n) = ( I(ps1,ns1) / I(ps2,ns2) ) * Scale
--
--              The sense terminals ps1 and ns1 behave as a short circuit.
--              The sense terminals ps2 and ns2 behave as a short circuit.
--
-- TERMINALS:
--
--  p           positive current source terminal
--  n           negative current source terminal
--  ps1         positive sense terminal for control current 1
--  ns1         negative sense terminal for control current 1
--  ps2         positive sense terminal for control current 2
--  ns2         negative sense terminal for control current 2
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output current
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
-- =========================================================
  entity IBIS_CCCS_DIV is

    generic ( 
      Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_CCCS_DIV;
  --===========================================================================
  architecture ideal of IBIS_CCCS_DIV is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across Iin1 through ps1 to ns1;
  quantity Vin2 across Iin2 through ps2 to ns2;

  begin

    Vin1 == 0.0;
    Vin2 == 0.0;
    if Iin2 /= 0.0 use
      Iout == Scale * Iin1 / Iin2;
    else
      Iout == real'high;     -- instead of inf;
    end use;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_VCCS_PWL (p, n, ps, ns)
--
-- DESCRIPTION:
--              A voltage-controlled current source with non-linear
--              relationship between control voltage and output current,
--              described using a piecewise-linear table.
--
--              Each voltage value in the array of X values is paired with
--              a Y array current value. The Length parameter must be set
--              to the number of elements in these arrays.
--
--              The input control voltage determines a position in the X
--              array, and the corersponding Y array value is used for the
--              output current. Linear interpolation is used for control
--              voltage values that do not fall exactly on a particular X
--              array value. Points outside the X array range are
--              extrapolated using the slope of the nearest two points of
--              the PWL data.
--
--              The sense terminals ps and ns draw no current.
--
-- TERMINALS:
--
--  p           positive current source terminal
--  n           negative current source terminal
--  ps          positive sense terminal for control voltage
--  ns          negative sense terminal for control voltage
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output current
--              DEFAULT Scale = '1.0'
--
--  Length      Number of elements in the X and Y arrays
--              DEFAULT Length = '4'
--
--  X[1:Length]  Array of X axis values (input control voltage)
--              DEFAULT X[1:Length] = '{-5.00 0.00 5.00 10.00}'
--
--  Y[1:Length]  Array of Y axis values (output current)
--              DEFAULT Y[1:Length] = '{-0.10 0.00 0.10 0.10}'
--
--
-- INSTANCE CALL EXAMPLES:
--
-- =========================================================
  entity IBIS_VCCS_PWL is

    generic ( 
      Scale  : real    := 1.0;
      -------------------------------------------------------------------------
      -- Vectors of the PWL table
      -------------------------------------------------------------------------
      X : real_vector := (-5.00,  0.00,  5.00, 10.00);
      Y : real_vector := (-0.10,  0.00,  0.10,  0.10));

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_VCCS_PWL;
  --===========================================================================
  architecture ideal of IBIS_VCCS_PWL is

  quantity Vout across Iout through p  to n;
  quantity Vin  across              ps to ns;
  --===========================================================================
  function Lookup (X             : in real;
                   Xdata         : in real_vector;
                   Ydata         : in real_vector;
                   Extrapolation : in string := "SE") return real is 
  -----------------------------------------------------------------------------
  -- This function returns "Y" that corresponds to "X" in the "Ydata" "Xdata"
  -- input pair using linear interpolation.
  --
  -- If the "X" input value lies outside the range of "Xdata", the returned
  -- "Y" value will either be equal to the first or last point in "Ydata",
  -- or it will be calculated using the slope between the first or last two
  -- points of "Ydata".  The extrapolation method is determined by the string
  -- in "Extrapolation".  "HE" stands for "Horizontal Extrapolation" which
  -- selects the former method, and "SE" stands for "Slope Extrapolation" which
  -- selects the latter method.
  --
  -- Acknowledgements:  The original version of this function was developed by
  -- Mentor Graphics Corporation.  Modifications added by Intel Corporation.
  -----------------------------------------------------------------------------
    variable xvalue, yvalue, m : real;
    variable start, fin, mid   : integer; 
  -----------------------------------------------------------------------------
  begin
    ---------------------------------------------------------------------------
    -- Handle cases when "X" is outside the range of "Xdata"
    ---------------------------------------------------------------------------
    if (Extrapolation = "SE") and (X <= Xdata(0)) then
      m := (Ydata(1) - Ydata(0)) / (Xdata(1) - Xdata(0));
      yvalue := Ydata(0) + m * (X - Xdata(0));
      return yvalue;

    elsif (Extrapolation = "HE") and (X <= Xdata(0)) then
      yvalue := Ydata(0);
      return yvalue;

    elsif (Extrapolation = "SE") and (X >= Xdata(Xdata'right)) then
      m := (Ydata(Ydata'right) - Ydata(Ydata'right - 1)) / (Xdata(Xdata'right) - Xdata(Xdata'right - 1));
      yvalue := Ydata(Ydata'right) + m * (X - Xdata(Xdata'right));
      return yvalue;

    elsif (Extrapolation = "HE") and (X >= Xdata(Xdata'right)) then
      yvalue := Ydata(Ydata'right);
      return yvalue;
    ---------------------------------------------------------------------------
    -- Handle cases when "X" is in the range of "Xdata"
    ---------------------------------------------------------------------------
    else
      start:= 0;
      fin := Xdata'right;

      while  start <= fin  loop
        mid := (start + fin) / 2; 

        if Xdata(mid) < X then
          start := mid + 1;
        else fin := mid - 1;
        end if;  

      end loop; 
                       
      if Xdata(mid) > X then
        mid := mid - 1; 
      end if;
      -------------------------------------------------------------------------
      -- Find "Y" by linear interpolation
      -------------------------------------------------------------------------
      yvalue := Ydata(mid) + (X - Xdata(mid)) * (Ydata(mid+1) - Ydata(mid)) / (Xdata(mid+1) - Xdata(mid));
      return yvalue;
    ---------------------------------------------------------------------------
    end if;
    ---------------------------------------------------------------------------
  end function Lookup;
  --===========================================================================

  begin

    Iout == Scale * Lookup(Vin, X, Y, "SE");

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_CCCS_PWL (p, n, ps, ns)
--
-- DESCRIPTION:
--              A current-controlled current source with non-linear
--              relationship between control current and output current,
--              described using a piecewise-linear table.
--
--              Each current value in the array of X values is paired with
--              a Y array current value. The Length parameter must be set
--              to the number of elements in these arrays.
--
--              The input control current determines a position in the X
--              array, and the corersponding Y array value is used for the
--              output current. Linear interpolation is used for control
--              current values that do not fall exactly on a particular X
--              array value. Points outside the X array range are
--              extrapolated using the slope of the nearest two points of
--              the PWL data.
--
--              The sense terminals ps and ns behave as a short circuit.
--
-- TERMINALS:
--
--  p           positive current source terminal
--  n           negative current source terminal
--  ps          positive sense terminal for control current
--  ns          negative sense terminal for control current
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output current
--              DEFAULT Scale = '1.0'
--
--  Length      Number of elements in the X and Y arrays
--              DEFAULT Length = '4'
--
--  X[1:Length]  Array of X axis values (input control current)
--              DEFAULT X[1:Length] = '{-5.00 0.00 5.00 10.00}'
--
--  Y[1:Length]  Array of Y axis values (output current)
--              DEFAULT Y[1:Length] = '{-0.10 0.00 0.10 0.10}'
--
--
-- INSTANCE CALL EXAMPLES:
--
-- =========================================================
  entity IBIS_CCCS_PWL is

    generic ( 
      Scale  : real    := 1.0;
      -------------------------------------------------------------------------
      -- Vectors of the PWL table
      -------------------------------------------------------------------------
      X : real_vector := (-5.00,  0.00,  5.00, 10.00);
      Y : real_vector := (-0.10,  0.00,  0.10,  0.10));

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_CCCS_PWL;
  --===========================================================================
  architecture ideal of IBIS_CCCS_PWL is

  quantity Vout across Iout through p  to n;
  quantity Vin  across Iin  through ps to ns;
  --===========================================================================
  function Lookup (X             : in real;
                   Xdata         : in real_vector;
                   Ydata         : in real_vector;
                   Extrapolation : in string := "SE") return real is 
  -----------------------------------------------------------------------------
  -- This function returns "Y" that corresponds to "X" in the "Ydata" "Xdata"
  -- input pair using linear interpolation.
  --
  -- If the "X" input value lies outside the range of "Xdata", the returned
  -- "Y" value will either be equal to the first or last point in "Ydata",
  -- or it will be calculated using the slope between the first or last two
  -- points of "Ydata".  The extrapolation method is determined by the string
  -- in "Extrapolation".  "HE" stands for "Horizontal Extrapolation" which
  -- selects the former method, and "SE" stands for "Slope Extrapolation" which
  -- selects the latter method.
  --
  -- Acknowledgements:  The original version of this function was developed by
  -- Mentor Graphics Corporation.  Modifications added by Intel Corporation.
  -----------------------------------------------------------------------------
    variable xvalue, yvalue, m : real;
    variable start, fin, mid   : integer; 
  -----------------------------------------------------------------------------
  begin
    ---------------------------------------------------------------------------
    -- Handle cases when "X" is outside the range of "Xdata"
    ---------------------------------------------------------------------------
    if (Extrapolation = "SE") and (X <= Xdata(0)) then
      m := (Ydata(1) - Ydata(0)) / (Xdata(1) - Xdata(0));
      yvalue := Ydata(0) + m * (X - Xdata(0));
      return yvalue;

    elsif (Extrapolation = "HE") and (X <= Xdata(0)) then
      yvalue := Ydata(0);
      return yvalue;

    elsif (Extrapolation = "SE") and (X >= Xdata(Xdata'right)) then
      m := (Ydata(Ydata'right) - Ydata(Ydata'right - 1)) / (Xdata(Xdata'right) - Xdata(Xdata'right - 1));
      yvalue := Ydata(Ydata'right) + m * (X - Xdata(Xdata'right));
      return yvalue;

    elsif (Extrapolation = "HE") and (X >= Xdata(Xdata'right)) then
      yvalue := Ydata(Ydata'right);
      return yvalue;
    ---------------------------------------------------------------------------
    -- Handle cases when "X" is in the range of "Xdata"
    ---------------------------------------------------------------------------
    else
      start:= 0;
      fin := Xdata'right;

      while  start <= fin  loop
        mid := (start + fin) / 2; 

        if Xdata(mid) < X then
          start := mid + 1;
        else fin := mid - 1;
        end if;  

      end loop; 
                       
      if Xdata(mid) > X then
        mid := mid - 1; 
      end if;
      -------------------------------------------------------------------------
      -- Find "Y" by linear interpolation
      -------------------------------------------------------------------------
      yvalue := Ydata(mid) + (X - Xdata(mid)) * (Ydata(mid+1) - Ydata(mid)) / (Xdata(mid+1) - Xdata(mid));
      return yvalue;
    ---------------------------------------------------------------------------
    end if;
    ---------------------------------------------------------------------------
  end function Lookup;
  --===========================================================================

  begin

    Vin  == 0.0;
    Iout == Scale * Lookup(Iin, X, Y, "SE");

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_TCCS_PWL (p, n)
--
-- DESCRIPTION:
--              A time-controlled current source with non-linear
--              relationship between time and output current, described
--              using a piecewise-linear table.
--
--              Each time value in the array of X values is paired with a
--              Y array votage value. The Length parameter must be set to
--              the number of elements in these arrays.
--
--              The arrays describe the output current waveform. The first
--              X value in the array gives the time at wich the waveform
--              begins in the simulation. Linear interpolation is used for
--              time values that do not fall exactly on a particular X
--              array value. Points outside the time range are
--              extrapolated horizontally, ie., the Y value of nearest
--              point is used.
--
-- TERMINALS:
--
--  p           positive current source terminal
--  n           negative current source terminal
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output current
--              DEFAULT Scale = '1.0'
--
--  Length      Number of elements in the X and Y arrays
--              DEFAULT Length = '4'
--
--  X[1:Length]  Array of X axis values (time)
--              DEFAULT X[1:Length] = '{0.50e-9 1.00e-9 2.00e-9 2.50e-9}'
--
--  Y[1:Length]  Array of Y axis values (output current)
--              DEFAULT Y[1:Length] = '{0.00 0.10 0.90 1.00}'
--
--
-- INSTANCE CALL EXAMPLES:
--
-- =========================================================
  entity IBIS_TCCS_PWL is

    generic ( 
      Scale : real := 1.0;
      -------------------------------------------------------------------------
      -- Vectors of the PWL table
      -------------------------------------------------------------------------
      X : real_vector := (0.50e-9,  1.00e-9,  2.00e-9,  2.50e-9);
      Y : real_vector := (0.00,     0.10,     0.90,     1.00));

    port (terminal p, n : electrical);

  end entity IBIS_TCCS_PWL;
  --===========================================================================
  architecture ideal of IBIS_TCCS_PWL is

  quantity Vout across Iout through p to n;
  --===========================================================================
  function Lookup (X             : in real;
                   Xdata         : in real_vector;
                   Ydata         : in real_vector;
                   Extrapolation : in string := "SE") return real is 
  -----------------------------------------------------------------------------
  -- This function returns "Y" that corresponds to "X" in the "Ydata" "Xdata"
  -- input pair using linear interpolation.
  --
  -- If the "X" input value lies outside the range of "Xdata", the returned
  -- "Y" value will either be equal to the first or last point in "Ydata",
  -- or it will be calculated using the slope between the first or last two
  -- points of "Ydata".  The extrapolation method is determined by the string
  -- in "Extrapolation".  "HE" stands for "Horizontal Extrapolation" which
  -- selects the former method, and "SE" stands for "Slope Extrapolation" which
  -- selects the latter method.
  --
  -- Acknowledgements:  The original version of this function was developed by
  -- Mentor Graphics Corporation.  Modifications added by Intel Corporation.
  -----------------------------------------------------------------------------
    variable xvalue, yvalue, m : real;
    variable start, fin, mid   : integer; 
  -----------------------------------------------------------------------------
  begin
    ---------------------------------------------------------------------------
    -- Handle cases when "X" is outside the range of "Xdata"
    ---------------------------------------------------------------------------
    if (Extrapolation = "SE") and (X <= Xdata(0)) then
      m := (Ydata(1) - Ydata(0)) / (Xdata(1) - Xdata(0));
      yvalue := Ydata(0) + m * (X - Xdata(0));
      return yvalue;

    elsif (Extrapolation = "HE") and (X <= Xdata(0)) then
      yvalue := Ydata(0);
      return yvalue;

    elsif (Extrapolation = "SE") and (X >= Xdata(Xdata'right)) then
      m := (Ydata(Ydata'right) - Ydata(Ydata'right - 1)) / (Xdata(Xdata'right) - Xdata(Xdata'right - 1));
      yvalue := Ydata(Ydata'right) + m * (X - Xdata(Xdata'right));
      return yvalue;

    elsif (Extrapolation = "HE") and (X >= Xdata(Xdata'right)) then
      yvalue := Ydata(Ydata'right);
      return yvalue;
    ---------------------------------------------------------------------------
    -- Handle cases when "X" is in the range of "Xdata"
    ---------------------------------------------------------------------------
    else
      start:= 0;
      fin := Xdata'right;

      while  start <= fin  loop
        mid := (start + fin) / 2; 

        if Xdata(mid) < X then
          start := mid + 1;
        else fin := mid - 1;
        end if;  

      end loop; 
                       
      if Xdata(mid) > X then
        mid := mid - 1; 
      end if;
      -------------------------------------------------------------------------
      -- Find "Y" by linear interpolation
      -------------------------------------------------------------------------
      yvalue := Ydata(mid) + (X - Xdata(mid)) * (Ydata(mid+1) - Ydata(mid)) / (Xdata(mid+1) - Xdata(mid));
      return yvalue;
    ---------------------------------------------------------------------------
    end if;
    ---------------------------------------------------------------------------
  end function Lookup;
  --===========================================================================

  begin

    Iout == Scale * Lookup(now, X, Y, "HE");

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_VECCS_PWL (p, n, ps, ns)
--
-- DESCRIPTION:
--              A voltage event-controlled current source producing
--              time-based piecewise-linear current waveforms in response
--              to trigger events. The control voltage triggers separate
--              rising and falling PWL waveforms to effect transitions
--              between high and low states.
--
--              Each time value in the XR array is paired with a current
--              value in the YR array to describe a rising edge waveform.
--              Likewise for XF and YF arrays, which describe a falling
--              edge waveform. The Length_R and Length_F parameters must
--              be set to the number of elements in the *R and *F arrays,
--              respectively.
--
--              When the control voltage rises above Vth_R the rising edge
--              waveform is output. Voltage then remains at the level of
--              the last point in the rising waveform table until a
--              falling edge transition is triggerd. When the control
--              voltage falls below Vth_F a falling edge waveform is
--              triggered, holding at the last current until the next
--              rising edge.
--
--              In DC analysis the first point of the falling edge
--              waveform is output if the control voltage is above Vth_R,
--              and the first point of the rising edge waveform is output
--              if the control voltage is below Vth_F. If Vt_R is less
--              than or equal to Vth_F and the control voltage lies on or
--              between those values the high state current is output. If
--              Vth_R is greater than Vth_F and the control voltage lies
--              on or between those values, the average of the high and
--              low state currents is output.
--
-- TERMINALS:
--
--  p           positive current source terminal
--  n           negative current source terminal
--  ps          positive sense terminal for control voltage
--  ns          negative sense terminal for control voltage
--
-- PARAMETERS:
--
--  Vth_R       Threshold voltage to trigger a rising edge event
--              DEFAULT Vth_R = '0.0'
--
--  Vth_F       Threshold voltage to trigger a falling edge event
--              DEFAULT Vth_F = '0.0'
--
--  Scale       Scaling factor for calculating output current
--              DEFAULT Scale = '1.0'
--
--  Length_R    Number of elements in the XR and YR arrays
--              DEFAULT Length_R = '4'
--
--  Length_F    Number of elements in the XF and YF arrays
--              DEFAULT Length_F = '4'
--
--  XR[1:Length_R]  Array of rising edge X values (time)
--              DEFAULT XR[1:Length_R] = '{0.50e-9 1.00e-9 2.00e-9 2.50e-9}'
--
--  YR[1:Length_R]  Array of rising edge Y values (current)
--              DEFAULT YR[1:Length_R] = '{0.00 0.10 0.90 1.00}'
--
--  XF[1:Length_F]  Array of falling edge X values (time)
--              DEFAULT XF[1:Length_F] = '{0.50e-9 1.00e-9 2.00e-9 2.50e-9}'
--
--  YF[1:Length_F]  Array of falling edge Y values (current)
--              DEFAULT YF[1:Length_F] = '{1.00 0.90 0.10 0.00}'
--
--
-- INSTANCE CALL EXAMPLES:
--
-- =========================================================
  entity IBIS_VECCS_PWL is

    generic ( 
      Vth_R  : real    := 0.0;
      Vth_F  : real    := 0.0;
      Scale  : real    := 1.0;
      -------------------------------------------------------------------------
      -- Vectors of the PWL table
      -------------------------------------------------------------------------
      XR : real_vector := (0.50e-9,  1.00e-9,  2.00e-9,  2.50e-9);
      YR : real_vector := (0.00,     0.10,     0.90,     1.00);
      XF : real_vector := (0.50e-9,  1.00e-9,  2.00e-9,  2.50e-9);
      YF : real_vector := (1.00,     0.90,     0.10,     0.00));

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_VECCS_PWL;
  --===========================================================================
  architecture ideal of IBIS_VECCS_PWL is

  quantity Vout across Iout through p  to n;
  quantity Vin  across              ps to ns;
  signal   T_event    : real    := 0.0;
  signal   EventState : integer := 0;
  --===========================================================================
  function Lookup (X             : in real;
                   Xdata         : in real_vector;
                   Ydata         : in real_vector;
                   Extrapolation : in string := "SE") return real is 
  -----------------------------------------------------------------------------
  -- This function returns "Y" that corresponds to "X" in the "Ydata" "Xdata"
  -- input pair using linear interpolation.
  --
  -- If the "X" input value lies outside the range of "Xdata", the returned
  -- "Y" value will either be equal to the first or last point in "Ydata",
  -- or it will be calculated using the slope between the first or last two
  -- points of "Ydata".  The extrapolation method is determined by the string
  -- in "Extrapolation".  "HE" stands for "Horizontal Extrapolation" which
  -- selects the former method, and "SE" stands for "Slope Extrapolation" which
  -- selects the latter method.
  --
  -- Acknowledgements:  The original version of this function was developed by
  -- Mentor Graphics Corporation.  Modifications added by Intel Corporation.
  -----------------------------------------------------------------------------
    variable xvalue, yvalue, m : real;
    variable start, fin, mid   : integer; 
  -----------------------------------------------------------------------------
  begin
    ---------------------------------------------------------------------------
    -- Handle cases when "X" is outside the range of "Xdata"
    ---------------------------------------------------------------------------
    if (Extrapolation = "SE") and (X <= Xdata(0)) then
      m := (Ydata(1) - Ydata(0)) / (Xdata(1) - Xdata(0));
      yvalue := Ydata(0) + m * (X - Xdata(0));
      return yvalue;

    elsif (Extrapolation = "HE") and (X <= Xdata(0)) then
      yvalue := Ydata(0);
      return yvalue;

    elsif (Extrapolation = "SE") and (X >= Xdata(Xdata'right)) then
      m := (Ydata(Ydata'right) - Ydata(Ydata'right - 1)) / (Xdata(Xdata'right) - Xdata(Xdata'right - 1));
      yvalue := Ydata(Ydata'right) + m * (X - Xdata(Xdata'right));
      return yvalue;

    elsif (Extrapolation = "HE") and (X >= Xdata(Xdata'right)) then
      yvalue := Ydata(Ydata'right);
      return yvalue;
    ---------------------------------------------------------------------------
    -- Handle cases when "X" is in the range of "Xdata"
    ---------------------------------------------------------------------------
    else
      start:= 0;
      fin := Xdata'right;

      while  start <= fin  loop
        mid := (start + fin) / 2; 

        if Xdata(mid) < X then
          start := mid + 1;
        else fin := mid - 1;
        end if;  

      end loop; 
                       
      if Xdata(mid) > X then
        mid := mid - 1; 
      end if;
      -------------------------------------------------------------------------
      -- Find "Y" by linear interpolation
      -------------------------------------------------------------------------
      yvalue := Ydata(mid) + (X - Xdata(mid)) * (Ydata(mid+1) - Ydata(mid)) / (Xdata(mid+1) - Xdata(mid));
      return yvalue;
    ---------------------------------------------------------------------------
    end if;
    ---------------------------------------------------------------------------
  end function Lookup;
  --===========================================================================

  begin

    ---------------------------------------------------------------------------
    Threshold_crossing : process is
    begin
      if domain = quiescent_domain then
        wait until domain /= quiescent_domain;
        if Vin > Vth_R then            -- high state
          EventState <= 1;
        elsif Vin < Vth_F then         -- low state
          EventState <= -1;
        end if;
      else
        wait on Vin'above(Vth_R), Vin'above(Vth_F);
        if Vin > Vth_R then            -- rising edge
          EventState <= 2;
          T_event    <= now;
        elsif Vin < Vth_F then         -- falling edge
          EventState <= -2;
          T_event    <= now;
        end if;
      end if;
    end process Threshold_crossing;
    ---------------------------------------------------------------------------

    case EventState use
      when -2 =>        -- Executes after falling edge events
        Iout == Scale * Lookup(now-T_event, XF, YF, "HE");
      when -1 =>        -- Executes between t=0 and first threshold crossing
        Iout == Scale * Lookup(0.0, XR, YR, "HE");    -- when input is LOW
      when 1 =>         -- Executes between t=0 and first threshold crossing
        Iout == Scale * Lookup(0.0, XF, YF, "HE");    -- when input is HIGH
      when 2 =>         -- Executes after rising edge events
        Iout == Scale * Lookup(now-T_event, XR, YR, "HE");
      when others =>    -- Executes during DC operating point analysis
        if Vin > Vth_R use
          Iout == Scale * Lookup(0.0, XF, YF, "HE");
        elsif Vin < Vth_F use
          Iout == Scale * Lookup(0.0, XR, YR, "HE");
        else
          Iout == Scale * (Lookup(0.0, XR, YR, "HE") + Lookup(0.0, XF, YF, "HE")) / 2.0;
        end use;
    end case;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_CECCS_PWL (p, n, ps, ns)
--
-- DESCRIPTION:
--              A current event-controlled current source producing
--              time-based piecewise-linear current waveforms in response
--              to trigger events. The control current triggers separate
--              rising and falling PWL waveforms to effect transitions
--              between high and low states.
--
--              Each time value in the XR array is paired with a current
--              value in the YR array to describe a rising edge waveform.
--              Likewise for XF and YF arrays, which describe a falling
--              edge waveform. The Length_R and Length_F parameters must
--              be set to the number of elements in the *R and *F arrays,
--              respectively.
--
--              When the control current rises above Ith_R the rising edge
--              waveform is output. Voltage then remains at the level of
--              the last point in the rising waveform table until a
--              falling edge transition is triggerd. When the control
--              current falls below Ith_F a falling edge waveform is
--              triggered, holding at the last current until the next
--              rising edge.
--
--              In DC analysis the first point of the falling edge
--              waveform is output if the control current is above Ith_R,
--              and the first point of the rising edge waveform is output
--              if the control current is below Ith_F. If It_R is less
--              than or equal to Ith_F and the control current lies on or
--              between those values the high state current is output. If
--              Ith_R is greater than Ith_F and the control current lies
--              on or between those values, the average of the high and
--              low state currents is output.
--
-- TERMINALS:
--
--  p           positive current source terminal
--  n           negative current source terminal
--  ps          positive sense terminal for control current
--  ns          negative sense terminal for control current
--
-- PARAMETERS:
--
--  Ith_R       Threshold current to trigger a rising edge event
--              DEFAULT Ith_R = '0.0'
--
--  Ith_F       Threshold current to trigger a falling edge event
--              DEFAULT Ith_F = '0.0'
--
--  Scale       Scaling factor for calculating output current
--              DEFAULT Scale = '1.0'
--
--  Length_R    Number of elements in the XR and YR arrays
--              DEFAULT Length_R = '4'
--
--  Length_F    Number of elements in the XF and YF arrays
--              DEFAULT Length_F = '4'
--
--  XR[1:Length_R]  Array of rising edge X values (time)
--              DEFAULT XR[1:Length_R] = '{0.50e-9 1.00e-9 2.00e-9 2.50e-9}'
--
--  YR[1:Length_R]  Array of rising edge Y values (current)
--              DEFAULT YR[1:Length_R] = '{0.00 0.10 0.90 1.00}'
--
--  XF[1:Length_F]  Array of falling edge X values (time)
--              DEFAULT XF[1:Length_F] = '{0.50e-9 1.00e-9 2.00e-9 2.50e-9}'
--
--  YF[1:Length_F]  Array of falling edge Y values (current)
--              DEFAULT YF[1:Length_F] = '{1.00 0.90 0.10 0.00}'
--
--
-- INSTANCE CALL EXAMPLES:
--
-- =========================================================
  entity IBIS_CECCS_PWL is

    generic ( 
      Ith_R  : real    := 0.0;
      Ith_F  : real    := 0.0;
      Scale  : real    := 1.0;
      -------------------------------------------------------------------------
      -- Vectors of the PWL table
      -------------------------------------------------------------------------
      XR : real_vector := (0.50e-9,  1.00e-9,  2.00e-9,  2.50e-9);
      YR : real_vector := (0.00,     0.10,     0.90,     1.00);
      XF : real_vector := (0.50e-9,  1.00e-9,  2.00e-9,  2.50e-9);
      YF : real_vector := (1.00,     0.90,     0.10,     0.00));

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_CECCS_PWL;
  --===========================================================================
  architecture ideal of IBIS_CECCS_PWL is

  quantity Vout across Iout through p  to n;
  quantity Vin  across Iin  through ps to ns;
  signal   T_event    : real    := 0.0;
  signal   EventState : integer := 0;
  --===========================================================================
  function Lookup (X             : in real;
                   Xdata         : in real_vector;
                   Ydata         : in real_vector;
                   Extrapolation : in string := "SE") return real is 
  -----------------------------------------------------------------------------
  -- This function returns "Y" that corresponds to "X" in the "Ydata" "Xdata"
  -- input pair using linear interpolation.
  --
  -- If the "X" input value lies outside the range of "Xdata", the returned
  -- "Y" value will either be equal to the first or last point in "Ydata",
  -- or it will be calculated using the slope between the first or last two
  -- points of "Ydata".  The extrapolation method is determined by the string
  -- in "Extrapolation".  "HE" stands for "Horizontal Extrapolation" which
  -- selects the former method, and "SE" stands for "Slope Extrapolation" which
  -- selects the latter method.
  --
  -- Acknowledgements:  The original version of this function was developed by
  -- Mentor Graphics Corporation.  Modifications added by Intel Corporation.
  -----------------------------------------------------------------------------
    variable xvalue, yvalue, m : real;
    variable start, fin, mid   : integer; 
  -----------------------------------------------------------------------------
  begin
    ---------------------------------------------------------------------------
    -- Handle cases when "X" is outside the range of "Xdata"
    ---------------------------------------------------------------------------
    if (Extrapolation = "SE") and (X <= Xdata(0)) then
      m := (Ydata(1) - Ydata(0)) / (Xdata(1) - Xdata(0));
      yvalue := Ydata(0) + m * (X - Xdata(0));
      return yvalue;

    elsif (Extrapolation = "HE") and (X <= Xdata(0)) then
      yvalue := Ydata(0);
      return yvalue;

    elsif (Extrapolation = "SE") and (X >= Xdata(Xdata'right)) then
      m := (Ydata(Ydata'right) - Ydata(Ydata'right - 1)) / (Xdata(Xdata'right) - Xdata(Xdata'right - 1));
      yvalue := Ydata(Ydata'right) + m * (X - Xdata(Xdata'right));
      return yvalue;

    elsif (Extrapolation = "HE") and (X >= Xdata(Xdata'right)) then
      yvalue := Ydata(Ydata'right);
      return yvalue;
    ---------------------------------------------------------------------------
    -- Handle cases when "X" is in the range of "Xdata"
    ---------------------------------------------------------------------------
    else
      start:= 0;
      fin := Xdata'right;

      while  start <= fin  loop
        mid := (start + fin) / 2; 

        if Xdata(mid) < X then
          start := mid + 1;
        else fin := mid - 1;
        end if;  

      end loop; 
                       
      if Xdata(mid) > X then
        mid := mid - 1; 
      end if;
      -------------------------------------------------------------------------
      -- Find "Y" by linear interpolation
      -------------------------------------------------------------------------
      yvalue := Ydata(mid) + (X - Xdata(mid)) * (Ydata(mid+1) - Ydata(mid)) / (Xdata(mid+1) - Xdata(mid));
      return yvalue;
    ---------------------------------------------------------------------------
    end if;
    ---------------------------------------------------------------------------
  end function Lookup;
  --===========================================================================

  begin

    ---------------------------------------------------------------------------
    Threshold_crossing : process is
    begin
      if domain = quiescent_domain then
        wait until domain /= quiescent_domain;
        if Iin > Ith_R then            -- high state
          EventState <= 1;
        elsif Iin < Ith_F then         -- low state
          EventState <= -1;
        end if;
      else
        wait on Iin'above(Ith_R), Iin'above(Ith_F);
        if Iin > Ith_R then            -- rising edge
          EventState <= 2;
          T_event    <= now;
        elsif Iin < Ith_F then         -- falling edge
          EventState <= -2;
          T_event    <= now;
        end if;
      end if;
    end process Threshold_crossing;
    ---------------------------------------------------------------------------

    Vin == 0.0;
    case EventState use
      when -2 =>        -- Executes after falling edge events
        Iout == Scale * Lookup(now-T_event, XF, YF, "HE");
      when -1 =>        -- Executes between t=0 and first threshold crossing
        Iout == Scale * Lookup(0.0, XR, YR, "HE");    -- when input is LOW
      when 1 =>         -- Executes between t=0 and first threshold crossing
        Iout == Scale * Lookup(0.0, XF, YF, "HE");    -- when input is HIGH
      when 2 =>         -- Executes after rising edge events
        Iout == Scale * Lookup(now-T_event, XR, YR, "HE");
      when others =>    -- Executes during DC operating point analysis
        if Iin > Ith_R use
          Iout == Scale * Lookup(0.0, XF, YF, "HE");
        elsif Iin < Ith_F use
          Iout == Scale * Lookup(0.0, XR, YR, "HE");
        else
          Iout == Scale * (Lookup(0.0, XR, YR, "HE") + Lookup(0.0, XF, YF, "HE")) / 2.0;
        end use;
    end case;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
-- =========================================moduledoc======
--
-- MODULE:      IBIS_T (n1, ref1, n2, ref2)
--
-- DESCRIPTION:
--              A lossless transmission line.
--
-- TERMINALS:
--
--  n1          Signal terminal for end 1
--  ref1        Reference terminal for end 1
--  n2          Signal terminal for end 2
--  ref2        Reference terminal for end 2
--
-- PARAMETERS:
--
--  Z0          Transmission line impedance in ohms
--              DEFAULT Z0 = '50.0'
--
--  TD          Time delay in seconds
--              DEFAULT TD = '1.0e-9'
--
--
-- INSTANCE CALL EXAMPLES:
--
-- =========================================================
  entity IBIS_T is

    generic ( 
      Z0 : real := 50.0;
      TD : real := 1.0e-9);

    port (terminal n1, ref1, n2, ref2 : electrical);

  end entity IBIS_T;
  --===========================================================================
  architecture Ideal of IBIS_T is

  quantity Vend1 across Iend1 through n1 to ref1;
  quantity Vend2 across Iend2 through n2 to ref2;

  quantity i12 : real;  -- Current wave traveling from end1 to end2
  quantity i21 : real;  -- Current wave traveling from end2 to end1

  begin

    Iend1 == i12 - i21'delayed(TD);
    Iend2 == i21 - i12'delayed(TD);
    Vend1 == Z0 * (2.0 * i12 - Iend1);
    Vend2 == Z0 * (2.0 * i21 - Iend2);

  end architecture Ideal;
--=============================================================================
--=============================================================================
